import numpy as np
from ..names import Names
from whitebox_workflows import Raster, WbEnvironment, RasterDataType, Vector, AttributeField, FieldDataType, FieldData
from ..folder_base import FolderBase
from ..vector_extension import VectorExtension
from ..raster_extension import RasterExtension
from .structure_attribute import StructureAttribute
import heapq
from dataclasses import dataclass, field
import math
from collections import defaultdict
import logging
import os

logger = logging.getLogger(__name__)

class Structure(FolderBase):    
    """
    Structure processing to get the outlet and modified structure and flow direction. 
    This apply to wetland, Feedlot, Dugout, Catchbasin and Manure Storage

    A structure will have a boundary polygon shapefile and an optional outlet point shapefile. 
    """

    #catchbasin and manure storage is not included here as they don't impact the flow direction
    structure_types_affection_flow_direction = ["wetland","feedlot","wascob","dugout"]

    def __init__(self, structure_type:str, 
                 output_folder:str, 
                 dem_raster:Raster, 
                 structure_polygon_vector:Vector, 
                 structure_polygon_vector_id_field_name: str = Names.field_name_id,                 
                 structure_outlet_point_vector:Vector = None, 
                 structure_outlet_point_vector_id_field_name:str = Names.field_name_id,
                 structure_area_threshold_ha = 0, 
                 structure_allow_multiple_outlets = False,
                 structure_using_PFO = True):
        super().__init__(output_folder)

        self.structure_type = structure_type
        self.__structure_polygon_vector = structure_polygon_vector
        self.__structure_polygon_vector_id_field_name = structure_polygon_vector_id_field_name

        self.__structure_outlet_point_vector = structure_outlet_point_vector
        self.__structure_outlet_point_vector_id_field_name = structure_outlet_point_vector_id_field_name
        
        self.__dem_raster = dem_raster

        if self.__structure_polygon_vector is None:
            raise ValueError(f"{self.structure_type}: The structure polygon shapefile is empty.")
        exist, self.__structure_polygon_vector_id_field_name = VectorExtension.check_field_in_vector(self.__structure_polygon_vector, self.__structure_polygon_vector_id_field_name)
        if not exist:
            raise ValueError(f"{self.structure_type}: The given id field {self.__structure_polygon_vector_id_field_name} doesn't exist in boundary shapefile.")

        if self.__structure_outlet_point_vector is not None:
            if self.__structure_outlet_point_vector_id_field_name is None:
                raise ValueError(f"{self.structure_type}: The outlet shapefile id field name is empty.")
            exist, self.__structure_outlet_point_vector_id_field_name = VectorExtension.check_field_in_vector(self.__structure_outlet_point_vector, self.__structure_outlet_point_vector_id_field_name)
            if not exist:
                raise ValueError(f"{self.structure_type}: The given id field {self.__structure_outlet_point_vector_id_field_name} doesn't exist in outlet shapefile.")

        #area threshold lower that which the structures will be removed.
        self.__structure_area_threshold_ha = structure_area_threshold_ha

        #parameters for Shawn's method
        self.__structure_allow_multiple_outlets = structure_allow_multiple_outlets
        self.__structure_using_PFO = structure_using_PFO
        
        #parameters for Zhangbin's method
        #number of sub-structures that one structure will be split.
        #default 3 for wetland
        #defautl 1 for all other structures
        self.__structure_split_max_num = 1

        #the threshold of flow acc that will be consiered as outlet
        self.__structure_acc_theshold = 0

        #output file name
        self.__structure_boundary_original_raster_name = f"{self.structure_type}BoundaryOriginal{Names.raster_extension}"
        self.__structure_boundary_original_vector_name = f"{self.structure_type}BoundaryOriginal{Names.shapefile_extension}"

        self.__structure_boundary_processed_raster_name = f"{self.structure_type}BoundaryProcessed{Names.raster_extension}"
        self.__structure_boundary_processed_vector_name = f"{self.structure_type}BoundaryProcessed{Names.shapefile_extension}"

        self.__structure_outlet_original_raster_name = f"{self.structure_type}OutletOriginal{Names.raster_extension}"
        self.__structure_outlet_processed_raster_name = f"{self.structure_type}OutletProcessed{Names.raster_extension}"        
        self.__structure_outlet_processed_vector_name = f"{self.structure_type}OutletProcessed{Names.shapefile_extension}"


        #individual structure list
        self.__attributes = None
#region properties

    @property
    def boundary_original_vector(self):
        if self.__structure_area_threshold_ha <=0:
            return self.__structure_polygon_vector
        
        vector = self.get_vector(self.__structure_boundary_original_vector_name)
        if vector is None:
            vector = self.wbe.filter_vector_features_by_area(
                    self.__structure_polygon_vector, self.__structure_area_threshold_ha * 10000)
            
            self.save_vector(vector, self.__structure_boundary_original_vector_name, True, True)
            self.get_vector(self.__structure_boundary_original_vector_name)
        return vector


    @property
    def boundary_original_raster(self):
        """Original boundary raster"""
        raster = self.get_raster(self.__structure_boundary_original_raster_name)
        
        if raster is None:  
            logger.info(f"Converting {self.structure_type} boundary from shapefile ({self.boundary_original_vector.file_name}) to raster ...")
            raster = VectorExtension.vector_polygons_to_raster_with_boarder(
                                                        polygon_vector=self.boundary_original_vector, 
                                                        base_raster = self.__dem_raster, 
                                                        field_name = self.__structure_polygon_vector_id_field_name)
            self.save_raster(raster, self.__structure_boundary_original_raster_name)

        return raster

    @property
    def boundary_processed_vector(self)->Vector:
        """
        Processed boundary vector. The only change is adding a subbasin column and removing the structure that are not in any of subbasins.
        The original fields are still there
        """
        return self.get_vector(self.__structure_boundary_processed_vector_name)            

    @property
    def boundary_processed_raster(self)->Raster:
        """Process boundary raster where the structures that are not in any of subbasins are removed."""
        return self.get_raster(self.__structure_boundary_processed_raster_name)
    
    @property
    def boundary_raster(self)->Raster:
        """Final boundary raster"""
        if self.boundary_processed_raster is None:
            return self.boundary_original_raster
        
        return self.boundary_processed_raster    
            
    @property
    def outlet_raster(self)->Raster:
        """Final outlet raster"""
        return self.outlet_processed_raster
    
    @property
    def outlet_vector(self)->Vector:
        """Final outlet vector"""
        vector = self.get_vector(self.__structure_outlet_processed_vector_name)

        if vector is None:
            logger.info(f"Converting {self.structure_type} outlet from raster to vector ...")
            vector = VectorExtension.add_id_for_raster_value(RasterExtension.raster_to_vector(self.outlet_raster, "point"))
            self.save_vector(vector, self.__structure_outlet_processed_vector_name)

        return vector
    
    @property
    def outlet_original_raster(self)->Raster:
        """Original outlet raster"""      
        raster = self.get_raster(self.__structure_outlet_original_raster_name)        
        if raster is None and self.__structure_outlet_point_vector is not None:            
            logger.info(f"Converting {self.structure_type} outlet from shapefile ({self.__structure_outlet_point_vector.file_name}) to raster ...")
            raster = self.wbe.vector_points_to_raster(self.__structure_outlet_point_vector, 
                                                      base_raster = self.__dem_raster, 
                                                      field_name = self.__structure_outlet_point_vector_id_field_name)
            self.save_raster(raster, self.__structure_outlet_original_raster_name)

        return raster

    @property
    def outlet_processed_raster(self):
        """Processed outlet raster"""
        raster = self.get_raster(self.__structure_outlet_processed_raster_name)        

        if raster is None:
            if self.outlet_original_raster is not None:                
                raster = self.__offset_outlets()
            else:                
                raster = self.__find_outlets_shawn()
            self.save_raster(raster,self.__structure_outlet_processed_raster_name)

        return raster
    
    @property 
    def attributes(self)->dict:
        """structure attribute dictionary including id, area, subbasin and contribution area"""
        if self.__attributes is None:
            if self.boundary_processed_vector is None:
                raise ValueError(f"Processed vector for structure: {self.structure_type} doesn't exist. ")

            self.__attributes = {}
            for i in range(self.boundary_processed_vector.num_records):
                id = int(self.boundary_processed_vector.get_attribute_value(i, Names.field_name_id).get_value_as_f64())
                subbasin = int(self.boundary_processed_vector.get_attribute_value(i, Names.field_name_subbasin).get_value_as_f64())
                area = self.boundary_processed_vector.get_attribute_value(i, Names.field_name_area).get_value_as_f64() / 10000.0
                contibution_area = self.boundary_processed_vector.get_attribute_value(i, Names.field_name_contibution_area_ha).get_value_as_f64()
                
                self.__attributes[id] = StructureAttribute(id, area, subbasin, contibution_area)
        
        return self.__attributes 

#endregion

#region 
    
    def __creatd_processed_boundary_vector(self, id_subbasin_dict:dict, id_contribution_area_dict:dict)->Vector:
        """Add subbasin column to vector with the given id subbasin dictionary"""
       
        #creat all necessary fields
        #we will just use four columns: id, subbasin, contribution area and area. Area will be added by polygon_area function.
        fields = [
            AttributeField(Names.field_name_id, FieldDataType.Int, 6, 0),
            AttributeField(Names.field_name_subbasin, FieldDataType.Int, 6, 0),
            AttributeField(Names.field_name_contibution_area_ha, FieldDataType.Real, 12, 5)
        ]

        out_vector = self.wbe.new_vector(self.boundary_original_vector.header.shape_type, 
                                         fields,
                                         proj=self.boundary_original_vector.projection)

        # Now fill it with the input data
        for i in range(self.boundary_original_vector.num_records):
            id = int(self.boundary_original_vector.get_attribute_value(i, self.__structure_polygon_vector_id_field_name).get_value_as_f64())
            
            #remove those that are not included in subbasin
            if id not in id_subbasin_dict:
                continue
            
            #add geometry
            geom = self.boundary_original_vector[i]
            out_vector.add_record(geom) # Add the record to the output Vector

            #add fields
            field_data = [
                FieldData.new_int(id),
                FieldData.new_int(int(id_subbasin_dict[id])),
                FieldData.new_real(id_contribution_area_dict[id])
            ]
            out_vector.add_attribute_record(field_data, deleted=False)

        #we also add the polygon area to the vector, the column name will be fixed as AREA
        return self.wbe.polygon_area(out_vector)

#endregion

    def repair_subbasin_and_assign_subbasin_id_contribution_area_to_sturcture(self, subbasin_raster:Raster, flow_acc_raster:Raster):
        """
        it modifies both subbasin and structure raster. 
        this has been used on wetland, feedlot and dugout

        replace repairSubbasinWithFlowDirChangeBMP
        """

        rows = self.boundary_raster.configs.rows
        cols = self.boundary_raster.configs.columns
        structure_raster_no_data = self.boundary_raster.configs.nodata
        subbasin_raster_no_data = subbasin_raster.configs.nodata
        mask_raster_no_data = self.__dem_raster.configs.nodata
        outlet_raster_node_data = self.outlet_raster.configs.nodata
        cell_size_ha = self.__dem_raster.configs.resolution_x * self.__dem_raster.configs.resolution_y / 10000.0

        #assuming 1:1 relationship from wetland to subbasin
        bmp2sub = {}
        bmp_contribute_area = {}
        for row in range(rows):
            for col in range(cols):
                wet_id = self.boundary_raster[row, col]
                sub_id = subbasin_raster[row, col]
                if wet_id != structure_raster_no_data and sub_id != subbasin_raster_no_data:
                    if int(wet_id) not in bmp2sub:
                        bmp2sub[int(wet_id)] = int(sub_id)

                    #get structure contribution area at the outlet    
                    if self.outlet_raster[row, col] != outlet_raster_node_data:
                        bmp_contribute_area[int(wet_id)] = flow_acc_raster[row, col] * cell_size_ha

        # Repair subbasin layer for the nodata cells within wetlands
        for row in range(rows):
            for col in range(cols):
                wet_id = self.boundary_raster[row, col]
                sub_id = subbasin_raster[row, col]
                if (sub_id == subbasin_raster_no_data and 
                    wet_id != structure_raster_no_data and 
                    self.__dem_raster[row, col] != mask_raster_no_data and 
                    int(wet_id) in bmp2sub):
                    subbasin_raster[row, col] = bmp2sub[int(wet_id)]

        # Remove wetlands that are not overlapping with subbasins
        for row in range(rows):
            for col in range(cols):
                if subbasin_raster[row, col] < 0:
                    self.boundary_raster[row, col] = structure_raster_no_data

        #save the change to file
        self.save_raster(self.boundary_raster, self.__structure_boundary_processed_raster_name)

        #assign subbasin id and contribution area to the shapefile also remove the structure that are not included in subbasin    
        processed_boundary_vector = self.__creatd_processed_boundary_vector(bmp2sub,bmp_contribute_area)
        self.save_vector(processed_boundary_vector, self.__structure_boundary_processed_vector_name)


#region Zhangbin - Not used anymore

    def __find_outlets_zhangbin(self):
        """
        1）首先找到湿地边界，寻找汇流累积量最大的3个cell，作为出口；
        对于其他流出的cell，强制流向湿地内；
        2）追溯每块湿地的上游
        :param structure_raster:  Raster of wetlands
        :param acc_threshold:   Threshold of accumulation,which is used to judge if the cell of extent can be the outlet

        Replace plugin WetlandOutletsPFO
        """ 
        
        #this is the ERSI pointer, different from the Whitebox pointer
        #dmove = [(0, 1), (1, 1), (1, 0), (1, -1), (0, -1), (-1, -1), (-1, 0), (-1, 1)]
        #dmove_dic = {1: (0, 1), 2: (1, 1), 4: (1, 0), 8: (1, -1), 16: (0, -1), 32: (-1, -1), 64: (-1, 0), 128: (-1, 1)}    

        #this is the whitebox direction
        dmove = [(-1, 1), (0, 1), (1, 1), (1, 0), (1, -1), (0, -1), (-1, -1), (-1, 0)]
        dmove_dic = {1: (-1, 1), 2: (0, 1), 4: (1, 1), 8: (1, 0), 16: (1, -1), 32: (0, -1), 64: (-1, -1), 128: (-1, 0)} 

        structure_raster = self.boundary_original_raster
        dem_raster = self.__dem_raster
        acc_threshold = self.__structure_acc_theshold
        area_threshold = self.__structure_area_threshold_ha * 10000

        #generate flow dir and flow accumulation
        flow_dir_raster = self.wbe.d8_pointer(dem = dem_raster)
        flow_acc_raster = self.wbe.d8_flow_accum(raster = flow_dir_raster, input_is_pointer = True)

        #some processing here
        wetland_nodata = structure_raster.configs.nodata
        cell_area = structure_raster.configs.resolution_x * structure_raster.configs.resolution_y
        row = structure_raster.configs.rows
        col = structure_raster.configs.columns

        Vis = np.zeros((row, col))
        Result = np.zeros((row, col))
        Result[:, :] = wetland_nodata
        Vis1 = np.zeros((row, col))
        Vis2 = np.zeros((row, col))
        id = 1
        wetland_extent = np.zeros((row, col))
        wetland_extent[:, :] = wetland_nodata  #边界标记为1
        OUTLET = np.zeros((row, col))
        OUTLET[:, :] = wetland_nodata

        WETLAND = np.zeros((row, col)) #重新编码后的wetland
        for i in range(row):
            for j in range(col):
                WETLAND[i,j] = structure_raster[i,j]

        out = np.zeros((row, col))
        out[:, :] = wetland_nodata

        removed_wetland_ids = []
        removed_wetland_ids.append(wetland_nodata)

        for i in range(row):
            for j in range(col):
                if structure_raster[i, j] != wetland_nodata:
                    # print(id)
                    now_wetland_id = structure_raster[i, j]
                    extent = [] #边界栅格和其对应的FLOW ACC
                    wetland_cells = []
                    Wetland_Area = 0
                    if Vis[i, j] == 0:
                        # 搜索邻域内湿地并标记边界栅格
                        pop_cells = [(i, j)]  # 迭代列表
                        Vis[i, j] = 1  # 标记已遍历
                        temp_A = [] #属于同一个structure的所有cell的row and col index
                        while pop_cells:
                            pop_cell = pop_cells.pop()
                            wetland_cells.append(pop_cell)
                            temp_A.append(pop_cell)
                            Wetland_Area += cell_area
                            # 搜索8邻域内cell
                            flag = False
                            for k in range(8):
                                next_cell = (pop_cell[0] + dmove[k][0], pop_cell[1] + dmove[k][1])
                                if 0 <= next_cell[0] < row and 0 <= next_cell[1] < col:  # 保证cell有效
                                    # if wetland[next_cell[0], next_cell[1]] != wetland[pop_cell[0], pop_cell[1]]:
                                    #     flag = True

                                    if Vis[next_cell[0], next_cell[1]] == 0 and structure_raster[next_cell[0], next_cell[1]] == \
                                            structure_raster[i, j]:
                                        pop_cells.append(next_cell)
                                        Vis[next_cell[0], next_cell[1]] = 1

                        # 判断是否为边界栅格
                        for cell in temp_A:
                            for k in range(8):
                                next_cell = (cell[0] + dmove[k][0], cell[1] + dmove[k][1])
                                if 0 <= next_cell[0] < row and 0 <= next_cell[1] < col:  # 保证cell有效
                                    if structure_raster[next_cell[0], next_cell[1]] != structure_raster[i, j]:
                                        extent.append((cell[0], cell[1], flow_acc_raster[cell[0], cell[1]]))
                                        break
                                else:
                                    extent.append((cell[0], cell[1], flow_acc_raster[cell[0], cell[1]]))
                                    break
                    for cell in extent:
                        wetland_extent[cell[0], cell[1]] = 1

                    # 判断是否为边界栅格
                    # 处理当前湿地
                    # 1、寻找出口
                    outlets = []

                    Method = 1
                    extent.sort(key=lambda x: x[2], reverse=True)  # 对汇流累积量排序  ！！！！！！！流出的最大！！！！！！！！！！！！！！！！！！
                    # print(extent)
                    temp_num = 0
                    for ii in range(len(extent)):  # 最多指定3个出口，小于阈值的不作为出口
                        if temp_num >= self.__structure_split_max_num:
                            break
                        # 先判断是否流出，再判断Acc阈值
                        now_dir = flow_dir_raster[extent[ii][0], extent[ii][1]]
                        if now_dir in dmove_dic:
                            #找到下游的cell
                            ds_cell = (extent[ii][0] + dmove_dic[now_dir][0], extent[ii][1] + dmove_dic[now_dir][1])
                            if 0 <= ds_cell[0] < row and 0 <= ds_cell[1] < col:
                                #向外流出的cell
                                if structure_raster[ds_cell[0], ds_cell[1]] != structure_raster[extent[ii][0], extent[ii][1]]:
                                    out[extent[ii][0], extent[ii][1]] = 2
                                    # 流向湿地外，判断Acc阈值
                                    # 只有超过一定的阙值才会作为出口
                                    if extent[ii][2] >= acc_threshold:
                                        #出口的位置和Acc
                                        outlets.append(extent[ii])

                                        #标记出口位置
                                        OUTLET[extent[ii][0], extent[ii][1]] = 1
                                        temp_num += 1
                    #去掉面积太小的
                    if Wetland_Area <= area_threshold:
                        removed_wetland_ids.append(now_wetland_id)
                        for cell in outlets[1:]:
                            outlets.remove(cell)
                            OUTLET[cell[0], cell[1]] = wetland_nodata

                        #if the wetland is removed, there is no need to process it further
                        continue

                    # 如果面积小于阈值则outlets==[]，此时说明湿地没有流向外面的cell，则以SEIMS方式处理
                    # print(outlets)
                    if len(outlets) == 0:
                        # print('**')
                        # outlets=extent.copy()
                        # outlets=wetland_cells.copy()
                        Method = 2

                    if Method == 2:
                        # SEIMS:按照湿地边界追溯上游
                        # 边界内的所有CELL
                        cells = wetland_cells.copy()
                        # print(cells)
                        # print(extent)
                        while cells:
                            cell = cells.pop()
                            Result[cell[0], cell[1]] = id #wetland重新编号
                            Us_cells = self.__get_upstream_cell(flow_dir_raster, cell[0], cell[1])
                            for temp_cell in Us_cells:
                                if Vis1[temp_cell[0], temp_cell[1]] == 0:
                                    if structure_raster[temp_cell[0], temp_cell[1]] == wetland_nodata:
                                        cells.append(temp_cell)
                                        Vis1[temp_cell[0], temp_cell[1]] = 1
                                        # Result[temp_cell[0], temp_cell[1]] = id
                        id += 1
                    if Method == 1:
                        # IMWEBS:先分割湿地，再回溯上游
                        # 2、强制边界cell流向湿地
                        pop_cells = outlets.copy()  # 迭代列表
                        # print(pop_cells)
                        for cell_1 in outlets:
                            wetland_extent[cell_1[0], cell_1[1]] = 2
                        while pop_cells:
                            pop_cell = pop_cells.pop()
                            # print(pop_cell)
                            for k in range(8):
                                next_cell_1 = (pop_cell[0] + dmove[k][0], pop_cell[1] + dmove[k][1])  # 临时变量
                                if 0 <= next_cell_1[0] < row and 0 <= next_cell_1[1] < col:  # 保证cell有效
                                    next_cell = (pop_cell[0] + dmove[k][0], pop_cell[1] + dmove[k][1],
                                                flow_acc_raster[next_cell_1[0], next_cell_1[1]])
                                    #just process the cells at the boundary
                                    if wetland_extent[next_cell[0], next_cell[1]] == 1:
                                        # 如果是下游是流出湿地，则纠正流向
                                        next_cell_dir = flow_dir_raster[next_cell[0], next_cell[1]]
                                        if next_cell_dir in dmove_dic:
                                            temp_next_cell = (next_cell[0] + dmove_dic[next_cell_dir][0],
                                                            next_cell[1] + dmove_dic[next_cell_dir][1])
                                            if 0 <= temp_next_cell[0] < row and 0 <= temp_next_cell[1] < col:
                                                if structure_raster[temp_next_cell[0], temp_next_cell[1]] != structure_raster[
                                                    next_cell[0], next_cell[1]]:
                                                    flow_dir_raster[next_cell[0], next_cell[1]] = 2 ** ((k + 4) % 8)
                                            wetland_extent[next_cell[0], next_cell[1]] = 2
                                            pop_cells.insert(0, next_cell)

                        # 3、追溯wetland上游并对分割后的wetland重新编码
                        for cell in outlets:
                            Vis9 = np.zeros((row, col))
                            cells = [cell]
                            Vis9[cell[0], cell[1]] = 1
                            while cells:
                                pop_cell = cells.pop()
                                Result[pop_cell[0], pop_cell[1]] = id
                                Us_cells = self.__get_upstream_cell(flow_dir_raster, pop_cell[0], pop_cell[1])
                                for temp_cell in Us_cells:
                                    if Vis9[temp_cell[0], temp_cell[1]] == 0:
                                        if structure_raster[temp_cell[0], temp_cell[1]] == wetland_nodata or structure_raster[
                                            temp_cell[0], temp_cell[1]] == now_wetland_id:
                                            cells.append(temp_cell)
                                            Vis9[temp_cell[0], temp_cell[1]] = 1
                            id += 1

        # 重构湿地
        Vis1[:, :] = 0
        second = {}
        for i in range(row):
            for j in range(col):
                if structure_raster[i, j] != wetland_nodata and structure_raster[i, j] not in removed_wetland_ids and Result[i, j] == wetland_nodata:
                    # 分割湿地后，有些湿地没有出口，需要再追溯上游
                    second.setdefault(structure_raster[i, j], []).append((i, j))

        for wet_ in second:
            wetland_cells = second[wet_]

            while wetland_cells:
                pop_cell = wetland_cells.pop()
                Result[pop_cell[0], pop_cell[1]] = id
                Us_cells = self.__get_upstream_cell(flow_dir_raster, pop_cell[0], pop_cell[1])
                for temp_cell in Us_cells:
                    if Vis1[temp_cell[0], temp_cell[1]] == 0:
                        if structure_raster[temp_cell[0], temp_cell[1]] == wetland_nodata or structure_raster[
                            temp_cell[0], temp_cell[1]] == wet_:
                            wetland_cells.append(temp_cell)
                            Vis1[temp_cell[0], temp_cell[1]] = 1
            id += 1

        # 重构编码
        us = {}
        for i in range(row):
            for j in range(col):
                if Result[i, j] != wetland_nodata:
                    us.setdefault(Result[i, j], []).append((i, j))
        
        #再一次重新编码，确保从1开始
        Result[:, :] = wetland_nodata
        id = 1
        for wet_ in us:
            wetland_cells = us[wet_]

            for cell in wetland_cells:
                Result[cell[0], cell[1]] = id
            id += 1

        for i in range(row):
            for j in range(col):
                if WETLAND[i, j] != wetland_nodata:
                    WETLAND[i, j] = Result[i, j]
                if OUTLET[i,j] != wetland_nodata:
                    OUTLET[i,j] = Result[i, j]

        #create rasters
        out_configs = structure_raster.configs
        out_configs.data_type = RasterDataType.F32

        wetland_extent_raster = self.wbe.new_raster(out_configs)
        wetland_upstream_raster = self.wbe.new_raster(out_configs)
        wetland_outlet_raster = self.wbe.new_raster(out_configs)
        wetland_modified_raster = self.wbe.new_raster(structure_raster.configs)
        wetland_out_raster = self.wbe.new_raster(out_configs)

        for i in range(row):
            for j in range(col):
                wetland_extent_raster[i,j] = wetland_extent[i,j]
                wetland_upstream_raster[i,j] = Result[i,j]
                wetland_outlet_raster[i,j] = OUTLET[i,j]
                wetland_modified_raster[i,j] = WETLAND[i,j]
                wetland_out_raster[i,j] = out[i,j]

        # 输出结果
        #return wetland_upstream_raster, wetland_extent_raster, wetland_outlet_raster, wetland_modified_raster, flow_dir_raster, wetland_out_raster
        return wetland_outlet_raster, wetland_modified_raster
    
    def __get_upstream_cell(self, flow_dir_raster:Raster, row:int, col:int):
        """
        查询输入栅格的上游栅格
        :param dir: array of dir
        :param row: row of the cell
        :param col:
        :return: [(i,j),(),]
        """
        #this is the ERSI pointer, different from the Whitebox pointer
        #dmove = [(0, 1), (1, 1), (1, 0), (1, -1), (0, -1), (-1, -1), (-1, 0), (-1, 1)]
        #dmove_dic = {1: (0, 1), 2: (1, 1), 4: (1, 0), 8: (1, -1), 16: (0, -1), 32: (-1, -1), 64: (-1, 0), 128: (-1, 1)}    

        #this is the whitebox direction
        dmove = [(-1, 1), (0, 1), (1, 1), (1, 0), (1, -1), (0, -1), (-1, -1), (-1, 0)]
        dmove_dic = {1: (-1, 1), 2: (0, 1), 4: (1, 1), 8: (1, 0), 16: (1, -1), 32: (0, -1), 64: (-1, -1), 128: (-1, 0)} 

        up_cell = []
        row_num = flow_dir_raster.configs.rows
        col_num = flow_dir_raster.configs.columns

        for i in range(8):
            now_loc = (row + dmove[i][0], col + dmove[i][1])
            # print(now_loc)
            if 0 <= now_loc[0]<row_num and 0 <= now_loc[1]<col_num:
                if flow_dir_raster[now_loc[0], now_loc[1]] == 2 ** ((i + 4) % 8):
                    up_cell.append(now_loc)

        return up_cell

    def __find_flow_path(self, structure_raster, structure_outlet_raster, flow_dir_raster, structure_upstream_raster):
        """
        Build the flow path within the wetlands.

        :param OutLet: Outlets of wetlands
        :param Dir: Flow direction
        :param Wetland: Raster of the wetlands
        :param LS: Output result, Upstream of the wetlands
        :param nodata: Nodata of the Wetland
        :return:
        """
        #this is the ERSI pointer, different from the Whitebox pointer
        #dmove = [(0, 1), (1, 1), (1, 0), (1, -1), (0, -1), (-1, -1), (-1, 0), (-1, 1)]
        #dmove_dic = {1: (0, 1), 2: (1, 1), 4: (1, 0), 8: (1, -1), 16: (0, -1), 32: (-1, -1), 64: (-1, 0), 128: (-1, 1)}    

        #this is the whitebox direction
        dmove = [(-1, 1), (0, 1), (1, 1), (1, 0), (1, -1), (0, -1), (-1, -1), (-1, 0)]
        dmove_dic = {1: (-1, 1), 2: (0, 1), 4: (1, 1), 8: (1, 0), 16: (1, -1), 32: (0, -1), 64: (-1, -1), 128: (-1, 0)} 

        row = structure_raster.configs.rows
        col = structure_raster.configs.columns
        wetland_nodata = structure_raster.configs.nodata

        Done = np.zeros((row, col))
        Vis = np.zeros((row, col))

        flow_tree = {}
        wbe = WbEnvironment()
        out_configs = structure_raster.configs
        out_configs.data_type = RasterDataType.I8
        out_configs.nodata = 0
        wetland_flowpath_raster = wbe.new_raster(structure_raster.configs)

        for i in range(row):
            for j in range(col):
                if structure_outlet_raster[i, j] != -9999:
                    # 找到出口，开始寻找下游
                    pop_cells = [(i, j)]
                    temp_road = [(i, j)]
                    # Vis[i, j] = 1
                    next_wetland_id = -1  # 没有下游湿地（流向边界外的或者洼地）为-1
                    now_wetland_id = structure_upstream_raster[i, j]
                    while pop_cells:
                        pop_cell = pop_cells.pop()
                        # print(pop_cell)
                        now_dir = flow_dir_raster[pop_cell[0], pop_cell[1]]
                        if now_dir in dmove_dic:
                            next_cell = (pop_cell[0] + dmove_dic[now_dir][0], pop_cell[1] + dmove_dic[now_dir][1])
                            if 0 <= next_cell[0] < row and 0 <= next_cell[1] < col:
                                if Vis[next_cell[0], next_cell[1]] == 0:
                                    if structure_raster[next_cell[0], next_cell[1]] != wetland_nodata and structure_raster[next_cell[0], next_cell[1]] != now_wetland_id:
                                        next_wetland_id = structure_upstream_raster[next_cell[0], next_cell[1]]
                                        if structure_outlet_raster[next_cell[0], next_cell[1]] != -9999:
                                            break
                                    pop_cells.append(next_cell)
                                    if next_cell in temp_road:
                                        break
                                    temp_road.append(next_cell)
                                    # Vis[next_cell[0], next_cell[1]] = 1
                    # 回溯路径
                    # print(temp_road)
                    for cell in temp_road:
                        Done[cell[0], cell[1]] = next_wetland_id
                        wetland_flowpath_raster[cell[0], cell[1]] = 1
                    flow_tree.setdefault(structure_raster[i, j], set()).add(next_wetland_id)

        return flow_tree, wetland_flowpath_raster    

    def generate_flow_direction_raster_zhangbin(self)->Raster:
        """
        change the flow direction so it flows to the outlet at each structure. 

        it assumes there is only one outlet for each structure. The id doesn't matter here.
        """

        #this is the ERSI pointer, different from the Whitebox pointer
        #dmove = [(0, 1), (1, 1), (1, 0), (1, -1), (0, -1), (-1, -1), (-1, 0), (-1, 1)]
        #dmove_dic = {1: (0, 1), 2: (1, 1), 4: (1, 0), 8: (1, -1), 16: (0, -1), 32: (-1, -1), 64: (-1, 0), 128: (-1, 1)}    

        #this is the whitebox direction
        dmove = [(-1, 1), (0, 1), (1, 1), (1, 0), (1, -1), (0, -1), (-1, -1), (-1, 0)]
        dmove_dic = {1: (-1, 1), 2: (0, 1), 4: (1, 1), 8: (1, 0), 16: (1, -1), 32: (0, -1), 64: (-1, -1), 128: (-1, 0)} 

        structure_raster = self.boundary_original_raster        
        outlet_raster = self.outlet_original_raster
        dem_raster = self.__dem_raster

        outlet_raster_no_data = outlet_raster.configs.nodata
   
        #generate original flow direction from dem
        flow_dir_raster = self.wbe.d8_pointer(dem = dem_raster)

        #get structure basic data
        wetland_nodata = structure_raster.configs.nodata
        cell_area = structure_raster.configs.resolution_x * structure_raster.configs.resolution_y
        row = structure_raster.configs.rows
        col = structure_raster.configs.columns

        Vis = np.zeros((row, col))              #search flag
        wetland_extent = np.zeros((row, col))
        wetland_extent[:, :] = wetland_nodata  #边界标记为1,出口标记为2

        for i in range(row):
            for j in range(col):
                if structure_raster[i, j] != wetland_nodata:
                    # print(id)
                    extent = [] #边界cell
                    wetland_cells = []
                    Wetland_Area = 0
                    outlets = []#出口cell,outlets is in extent

                    if Vis[i, j] == 0:
                        # 搜索邻域内湿地并标记边界栅格
                        pop_cells = [(i, j)]  # 迭代列表
                        Vis[i, j] = 1  # 标记已遍历
                        temp_A = [] #属于同一个structure的所有cell的row and col index
                        while pop_cells:
                            pop_cell = pop_cells.pop()
                            wetland_cells.append(pop_cell)
                            temp_A.append(pop_cell)
                            Wetland_Area += cell_area
                            # 搜索8邻域内cell
                            flag = False
                            for k in range(8):
                                next_cell = (pop_cell[0] + dmove[k][0], pop_cell[1] + dmove[k][1])
                                if 0 <= next_cell[0] < row and 0 <= next_cell[1] < col:  # 保证cell有效
                                    if Vis[next_cell[0], next_cell[1]] == 0 and structure_raster[next_cell[0], next_cell[1]] == \
                                            structure_raster[i, j]:
                                        pop_cells.append(next_cell)
                                        Vis[next_cell[0], next_cell[1]] = 1

                        # 判断是否为边界栅格
                        for cell in temp_A:
                            for k in range(8):
                                next_cell = (cell[0] + dmove[k][0], cell[1] + dmove[k][1])
                                if 0 <= next_cell[0] < row and 0 <= next_cell[1] < col:  # 保证cell有效
                                    if structure_raster[next_cell[0], next_cell[1]] != structure_raster[i, j]:
                                        extent.append((cell[0], cell[1]))

                                        if outlet_raster[cell[0], cell[1]] != outlet_raster_no_data:
                                            outlets.append(extent[-1])
                                        break
                                else:
                                    extent.append((cell[0], cell[1]))
                                    if outlet_raster[cell[0], cell[1]] != outlet_raster_no_data:
                                            outlets.append(extent[-1])
                                            
                                    break
                    for cell in extent:
                        wetland_extent[cell[0], cell[1]] = 1

                    if len(outlets) > 0:
                        pop_cells = outlets.copy()  # 迭代列表
                        # print(pop_cells)
                        for cell_1 in outlets:
                            wetland_extent[cell_1[0], cell_1[1]] = 2

                        #保证所有的边界cell都指向出口
                        while pop_cells:
                            pop_cell = pop_cells.pop()
                            # print(pop_cell)
                            for k in range(8):
                                next_cell_1 = (pop_cell[0] + dmove[k][0], pop_cell[1] + dmove[k][1])  # 临时变量
                                if 0 <= next_cell_1[0] < row and 0 <= next_cell_1[1] < col:  # 保证cell有效
                                    next_cell = (pop_cell[0] + dmove[k][0], pop_cell[1] + dmove[k][1])
                                    #just process the cells at the boundary
                                    if wetland_extent[next_cell[0], next_cell[1]] == 1:
                                        # 如果是下游是流出湿地，则纠正流向
                                        next_cell_dir = flow_dir_raster[next_cell[0], next_cell[1]]
                                        if next_cell_dir in dmove_dic:
                                            temp_next_cell = (next_cell[0] + dmove_dic[next_cell_dir][0],
                                                            next_cell[1] + dmove_dic[next_cell_dir][1])
                                            if 0 <= temp_next_cell[0] < row and 0 <= temp_next_cell[1] < col:
                                                if structure_raster[temp_next_cell[0], temp_next_cell[1]] != structure_raster[
                                                    next_cell[0], next_cell[1]]:
                                                    flow_dir_raster[next_cell[0], next_cell[1]] = 2 ** ((k + 4) % 8)#指向上一个边界cell或出口
                                            wetland_extent[next_cell[0], next_cell[1]] = 2
                                            pop_cells.insert(0, next_cell)



        wetland_boarder_raster = self.wbe.new_raster(structure_raster.configs)
        for i in range(row):
            for j in range(col):
                wetland_boarder_raster[i,j] = wetland_extent[i,j]
        self.save_raster(wetland_boarder_raster, f"{self.structure_type}BoarderLine{Names.raster_extension}")   


        return flow_dir_raster

#endregion Shawn

#region Shawn

    def __find_outlets_shawn(self):
        """
        Find structure outlets based using PFO methods

        Replace Plugin WetlandOutletsPFO        
        """

        logger.info(f"Looking for {self.structure_type} outlets ...")

        dem_raster = self.__dem_raster
        structure_raster = self.boundary_original_raster
        allow_multiple_output = self.__structure_allow_multiple_outlets
        using_PFO = self.__structure_using_PFO

        col, row, colN, rowN = 0, 0, 0, 0
        n = 0
        z, zN = 0.0, 0.0
        isPit, isEdgeCell, flag = False, False, False
        dX = [1, 1, 1, 0, -1, -1, -1, 0]
        dY = [-1, 0, 1, 1, 1, 0, -1, -1]
        backLink = [5, 6, 7, 8, 1, 2, 3, 4]

        nodata = dem_raster.configs.nodata
        rows = dem_raster.configs.rows
        cols = dem_raster.configs.columns

        dem2D =  np.full((rows, cols), nodata)
        output = np.full((rows, cols), nodata)
        flowdir = np.zeros((rows, cols), dtype=np.int8)
        queue = []
        inQueue = np.zeros((rows, cols), dtype=bool)

        flatIndex = 0
        k = 0
        dir = 0
        x, y = 0, 0

        if using_PFO:
            for row in range(rows):
                for col in range(cols):
                    z = dem_raster[row, col]
                    dem2D[row, col] = z
                    flowdir[row, col] = 0
                    if z != nodata:
                        isPit = True
                        isEdgeCell = False
                        for n in range(8):
                            zN = dem_raster[row + dY[n], col + dX[n]]
                            if isPit and zN != nodata and zN < z:
                                isPit = False
                            elif zN == nodata:
                                isEdgeCell = True
                        if isPit and isEdgeCell:                            
                            heapq.heappush(queue, GridCell(row=row, col=col, z=z, flatIndex = 0, streamVal=True))
                            inQueue[row, col] = True

            while queue:
                gc = heapq.heappop(queue)
                logger.debug(f"Row: {gc.row}, Col: {gc.col}, z: {gc.z}, flatIndex: {gc.flatIndex}")                
                z = gc.z
                row = gc.row
                col = gc.col    
                flatIndex = gc.flatIndex

                for n in range(8):
                    rowN = row + dY[n]
                    colN = col + dX[n]
                    if rowN >= rows or colN >= cols or rowN < 0 or colN < 0:
                        continue
                    zN = dem2D[rowN, colN]
                    if zN != nodata and not inQueue[rowN, colN]:
                        flowdir[rowN, colN] = backLink[n]
                        k = 0
                        if zN == dem2D[row, col]:
                            k = flatIndex + 1                            
                        heapq.heappush(queue, GridCell(row=rowN, col=colN, z=zN, flatIndex = k, streamVal=True))
                        inQueue[rowN, colN] = True

                        if structure_raster[row, col] != structure_raster[rowN, colN]:
                            flag = True
                            x = colN
                            y = rowN
                            isReturn = False
                            while flag:
                                #find the downslope neighbour
                                dir = flowdir[y, x] - 1
                                if dir > 0:
                                    if dir > 7:
                                        return
                                    x += dX[dir]
                                    y += dY[dir]
                                    if structure_raster[y, x] > 0:
                                        if structure_raster[rowN, colN] == structure_raster[y, x]:
                                            isReturn = True
                                            flag = False
                                        else:
                                            isReturn = False
                                            flag = False
                                else:
                                    flag = False
                            if not isReturn:
                                output[rowN, colN] = structure_raster[rowN, colN]
        else:
            c = 0
            flow_direction_raster = self.wbe.d8_pointer(dem = dem_raster)
            for row in range(rows):
                for col in range(cols):
                    if flow_direction_raster[row, col] > 0 and structure_raster[row, col] > 0:
                        z = dem_raster[row, col]
                        dem2D[row, col] = z
                        dir = int(flow_direction_raster[row, col])
                        if dir > 0:
                            c = int(math.log(dir) / self.LnOf2)
                            if c > 7:
                                raise ValueError("An unexpected value has been identified in the pointer image. This tool requires a pointer grid that has been created using either the D8 or Rho8 tools.")
                                
                            rowN = row + dX[c]
                            colN = col + dY[c]
                            if structure_raster[row, col] != structure_raster[rowN, colN]:
                                flag = True
                                x = colN
                                y = rowN
                                isReturn = False
                                count = 0
                                while flag and count < 100:
                                    dir = int(flow_direction_raster[y, x])
                                    if dir > 0:
                                        c = int(math.log(dir) / self.LnOf2)
                                        if c > 7:
                                            raise ValueError("An unexpected value has been identified in the pointer image. This tool requires a pointer grid that has been created using either the D8 or Rho8 tools.")

                                        if structure_raster[y, x] > 0:
                                            if structure_raster[row, col] == structure_raster[y, x]:
                                                isReturn = True
                                                flag = False
                                            else:
                                                isReturn = False
                                                flag = False
                                        x += dX[c]
                                        y += dY[c]
                                    else:
                                        flag = False
                                    count += 1
                                if not isReturn:
                                    output[row, col] = structure_raster[row, col]

        #Group the outlets and use the lowest one as the outlet
        outputNew = np.full((rows, cols), nodata)
        if allow_multiple_output:
            register = {}
            registerReverse = np.full((rows, cols), int(nodata))
            grpNum = 0
            for row in range(rows):
                for col in range(cols):
                    if output[row, col] > 0:
                        loc = [row, col]
                        num = -1
                        for n in range(8):
                            rowN = row + dY[n]
                            colN = col + dX[n]
                            if output[row, col] == output[rowN, colN]:
                                if registerReverse[rowN, colN] > 0:
                                    num = registerReverse[rowN, colN]
                                    break
                        if num < 0:
                            grpNum += 1
                            register[grpNum] = []
                            num = grpNum
                        registerReverse[row, col] = num
                        register[num].append(loc)

            for locs in register.values():
                elevMap = {}
                for loc in locs:
                    elevMap[dem_raster[loc[0], loc[1]]] = loc
                elev = list(elevMap.keys())
                elev.sort()
                lowest = elev[0]
                outputNew[elevMap[lowest][0], elevMap[lowest][1]] = output[elevMap[lowest][0], elevMap[lowest][1]]

            #For multiple outlets wetland, outlet flow to wetland will not be count
            wetlandOutlets = {}
            suspectCounts = {}
            suspects = []
            for row in range(rows):
                for col in range(cols):
                    if outputNew[row, col] > 0 and structure_raster[row, col] > 0:
                        id = int(structure_raster[row, col])
                        if id in wetlandOutlets:
                            wetlandOutlets[id] += 1
                        else:
                            wetlandOutlets[id] = 1
                        dir = flowdir[row, col] - 1
                        if dir > 0:
                            if dir > 7:
                                return
                            rowN = row + dX[dir]
                            colN = col + dY[dir]
                            if structure_raster[rowN, colN] > 0:
                                loc = [row, col]
                                suspects.append(loc)
                                if id in suspectCounts:
                                    suspectCounts[id] += 1
                                else:
                                    suspectCounts[id] = 1

            for loc in suspects:
                id = int(structure_raster[loc[0], loc[1]])
                if wetlandOutlets[id] > 1 and wetlandOutlets[id] > suspectCounts[id]:
                    outputNew[loc[0], loc[1]] = nodata
        else:
            register = {}
            for row in range(rows):
                for col in range(cols):
                    if output[row, col] > 0:
                        id = int(output[row, col])
                        loc = [row, col]
                        if id not in register:
                            register[id] = []
                        register[id].append(loc)

            for locs in register.values():
                elevMap = {}
                for loc in locs:
                    elevMap[dem_raster[loc[0], loc[1]]] = loc
                elev = list(elevMap.keys())
                elev.sort()
                lowest = elev[0]
                outputNew[elevMap[lowest][0], elevMap[lowest][1]] = output[elevMap[lowest][0], elevMap[lowest][1]]

        structure_outlet_raster = self.wbe.new_raster(dem_raster.configs)
        for row in range(rows):
            for col in range(cols):
                z = dem2D[row, col]
                if z != nodata:
                    structure_outlet_raster[row, col] = outputNew[row, col]

        return structure_outlet_raster

    def __categorize_wetland(self,
                main_stream_buffer_raster:Raster = None
                ):
        """
        1. Filter wetland with given min area, 
        2. Separate wetlands that have multiple outlets if user provides the outlets
        3. Separate isolated wetlands and riparian wetland if main stream buffer is provided
        4. Remove outlets that doesn't fall in the wetland boundary.

        Not used, instead just use polygon area to filter

        Replace filterWetland method.
        """

        if self.structure_type != "wetland":
            return None, None, None, None
        
        wetland_raster = self.boundary_original_raster
        wetland_outlet_raster = self.outlet_original_raster

        cell_area = wetland_raster.configs.resolution_x * wetland_raster.configs.resolution_y
        rows = wetland_raster.configs.rows
        cols = wetland_raster.configs.columns
        min_wetland_area_m2 = self.__structure_area_threshold_ha * 10000
        nodata = wetland_raster.configs.nodata

        wet_pixels = {}
        wet_outlet_count = {}
        rip_wet = set()

        for row in range(rows):
            for col in range(cols):
                if wetland_raster[row, col] > 0:
                    wet_id = int(wetland_raster[row, col])

                    #get the number of cells for each wetland
                    wet_pixels[wet_id] = wet_pixels.setdefault(wet_id, 0) + 1

                    #get the number of outlets for each wetland
                    #this may have issues as the outlet may not be in the wetland polygon right now
                    if wetland_outlet_raster is not None and wetland_outlet_raster[row, col] > 0:
                        wet_outlet_count[wet_id] = wet_outlet_count.setdefault(wet_id, 0) +  1
                    
                    #if the wetland is in the buffer of the stream, add to the riparian buffer list
                    if main_stream_buffer_raster is not None and main_stream_buffer_raster[row, col] > 0 and wet_id not in rip_wet:
                        rip_wet.add(wet_id)

        #remove the wetland that is smaller than given threshold value
        #this may has problem the rasterized wetland area may be smaller or larger than the polygon area depending on the resolution.
        #filtering by golygo area is much more accurate. 
        for wet_id in list(wet_pixels.keys()):
            if wet_pixels[wet_id] * cell_area < min_wetland_area_m2:
                wet_pixels[wet_id] = 0.0

        #wetlands that exceeds the threashold area
        wetland_no_small_raster = self.wbe.new_raster(wetland_raster.configs)

        #wetlands that have multiple outlets
        wetland_no_small_multi_raster = None
        if wetland_outlet_raster is not None:
            wetland_no_small_multi_raster = self.wbe.new_raster(wetland_raster.configs)
        
        #riparian and isolated wetlands
        wetland_isolated_raster = None
        wetland_rip_raster = None
        if main_stream_buffer_raster is not None:
            wetland_isolated_raster = self.wbe.new_raster(wetland_raster.configs)
            wetland_rip_raster = self.wbe.new_raster(wetland_raster.configs)

        for row in range(rows):
            for col in range(cols):
                wet_id = int(wetland_raster[row, col])
                if wet_id > 0 and wet_pixels[wet_id] > 0:
                    #skip the wetland if wetland doesn't have an outlet in the outlet raster
                    if wetland_outlet_raster is not None and wet_id not in wet_outlet_count:
                        continue
                    
                    wetland_no_small_raster[row, col] = wet_id

                    if wetland_outlet_raster is not None and wet_outlet_count[wet_id] > 1:
                        wetland_no_small_multi_raster[row, col] = wet_id

                    if main_stream_buffer_raster is not None:
                        if wet_id in rip_wet:
                            wetland_rip_raster[row, col] = wet_id
                        else:
                            wetland_isolated_raster[row, col] = wet_id

                elif wetland_outlet_raster and wetland_outlet_raster[row, col] > 0 and (wet_id not in wet_pixels or wet_pixels[wet_id] == 0):
                    #remove wetland outlets that are not inside of wetland boundary
                    wetland_outlet_raster[row, col] = nodata

        return wetland_no_small_raster, wetland_no_small_multi_raster, wetland_rip_raster, wetland_isolated_raster

    def generate_flow_direction_raster_shawn(self,
                                             stream_raster:Raster = None,
                                             lowerThresholdArea = 0,
                                             upperThresholdArea = float('inf'),
                                             reference_flow_direction_raster:Raster = None):
        """
        Generates the flow direction raster

        Replace plugin WetlandDirection
        """

        logger.info("Creating flow direction based on dem and structures ...")

        dem_raster = self.__dem_raster
        structure_vector = self.boundary_original_vector
        structure_raster = self.boundary_original_raster
        structure_outlet_raster = self.outlet_original_raster  

        #outputs
        structure_with_multiple_outlets_raster = None                                   #structures that have multiple outlets as the area exceeds the maximum area
        structure_riparian_raster = None                                                #structures that closes to stream
        structure_kept_raster = self.wbe.new_raster(dem_raster.configs)                 #structures that are kept for analysis the area is smaller than the maximum area but larger than the minimum area
        structure_isolated_raster = self.wbe.new_raster(dem_raster.configs)              #structure_with_multiple_outlets_raster + structure_with_single_outlet_raster
        structure_removed_raster = self.wbe.new_raster(dem_raster.configs)               #structures that remove as the area is smaller than minimum area
        flow_direction_raster = self.wbe.new_raster(dem_raster.configs)                  #final flow direction raster

        col, row, colN, rowN =  0, 0, 0, 0
        dir, n = 0, 0
        z, zN, zTest, zN2, lowestNeighbour = 0.0, 0.0, 0.0, 0.0, 0.0
        isPit, isEdgeCell, isWetland, flag = False, False, False, False
        gc = None

        # /*
        #  * 7  8  1
        #  * 6  X  2
        #  * 5  4  3
        #  */
        dX = [1, 1, 1, 0, -1, -1, -1, 0]
        dY = [-1, 0, 1, 1, 1, 0, -1, -1]
        backLink = [5, 6, 7, 8, 1, 2, 3, 4]
        outPointer = [0, 1, 2, 4, 8, 16, 32, 64, 128]
        inflowingVals = [16, 32, 64, 128, 1, 2, 4, 8]

        nodata = dem_raster.configs.nodata
        rows = dem_raster.configs.rows
        cols = dem_raster.configs.columns
        nodata_wetland = structure_raster.configs.nodata

        wetlandID = np.full((rows, cols), -32768, dtype=np.int32)
        wetlandPosition = np.full((rows, cols), -32768, dtype=np.int32)
        wetlandEndNodes = np.zeros((rows, cols), dtype=bool)

        count = 0
        points = []
        startingPointInPart, endingPointInPart = 0, 0
        i = 0
        x1, y1, x2, y2 = 0, 0, 0, 0


        conflictWetlands = set()
        wetland_boarder_raster = self.wbe.new_raster(dem_raster.configs) 

        #go through all the vertexs along the boundary to find confict wetlands and wetland position. 
        for index in range(structure_vector.num_records):
            #let's use wetland id here
            recNum = int(structure_vector.get_attribute_value(index, self.__structure_polygon_vector_id_field_name).get_value_as_f64())
            points = structure_vector[index].points
            numPoints = len(points)
            partData = structure_vector[index].parts
            numParts = len(partData)
            for part in range(numParts):
                startingPointInPart = partData[part]
                if part < numParts - 1:
                    endingPointInPart = partData[part + 1]
                else:
                    endingPointInPart = numPoints
                n = 0
                for i in range(startingPointInPart, endingPointInPart - 1):
                    x1 = dem_raster.get_column_from_x(points[i].x)
                    y1 = dem_raster.get_row_from_y(points[i].y)

                    x2 = dem_raster.get_column_from_x(points[i+1].x)
                    y2 = dem_raster.get_row_from_y(points[i+1].y)

                    d = 0

                    dy = abs(y2 - y1)
                    dx = abs(x2 - x1)

                    dy2 = dy << 1
                    dx2 = dx << 1

                    ix = 1 if x1 < x2 else -1
                    iy = 1 if y1 < y2 else -1

                    if dy <= dx:
                        while True:
                            if dem_raster[y1, x1] != nodata:
                                if wetland_boarder_raster[y1, x1] > 0 and wetland_boarder_raster[y1, x1] != recNum:
                                    wetID = int(structure_raster[y1, x1])
                                    if wetID > 0 and wetID not in conflictWetlands:
                                        conflictWetlands.add(wetID)
                                        for n in range(8):
                                            rowN = y1 + dY[n]
                                            colN = x1 + dX[n]
                                            neighbourID = int(structure_raster[rowN, colN])
                                            if neighbourID > 0 and neighbourID != wetID:
                                                if neighbourID not in conflictWetlands:
                                                    conflictWetlands.add(neighbourID)
                                else:
                                    n += 1
                                    wetland_boarder_raster[y1, x1] = recNum
                                    wetlandPosition[y1, x1] = n
                            if x1 == x2:
                                break
                            x1 += ix
                            d += dy2
                            if d > dx:
                                y1 += iy
                                d -= dx2
                    else:
                        while True:
                            if dem_raster[y1, x1] != nodata:
                                if wetland_boarder_raster[y1, x1] > 0 and wetland_boarder_raster[y1, x1] != recNum:
                                    wetID = int(structure_raster[y1, x1])
                                    if wetID > 0 and wetID not in conflictWetlands:
                                        conflictWetlands.add(wetID)
                                        for n in range(8):
                                            rowN = y1 + dY[n]
                                            colN = x1 + dX[n]
                                            neighbourID = int(structure_raster[rowN, colN])
                                            if neighbourID > 0 and neighbourID != wetID:
                                                if neighbourID not in conflictWetlands:
                                                    conflictWetlands.add(neighbourID)
                                else:
                                    n += 1
                                    wetland_boarder_raster[y1, x1] = recNum
                                    wetlandPosition[y1, x1] = n
                            if y1 == y2:
                                break
                            y1 += iy
                            d += dx2
                            if d > dy:
                                x1 += ix
                                d -= dy2
        self.save_raster(wetland_boarder_raster, f"{self.structure_type}BoarderLine{Names.raster_extension}")      

        wetland_position_raster = self.wbe.new_raster(dem_raster.configs)
        for row in range(rows):
            for col in range(cols):
                wetland_position_raster[row, col] = wetlandPosition[row, col]
        self.save_raster(wetland_position_raster, f"{self.structure_type}BoarderVertexIndex{Names.raster_extension}")      

        confWetGroups = defaultdict(list)
        confWetIsAdded = {id: -1 for id in conflictWetlands}

        riparianWetlandIDs = set()
        riparianWetlandRemoveIDs = set()
        if stream_raster is None:
            logger.info("No stream provided, no riparian wetland found!")

        wetlandCount = {}#numbe of cells of each wetland
        nodataWet = structure_raster.configs.nodata
        counter = 0
        for row in range(rows):
            for col in range(cols):
                if wetlandID[row, col] == -32768 and dem_raster[row, col] != nodata and structure_raster[row, col] > 0:
                    id = int(structure_raster[row, col])
                    wetlandPix = wetlandCount.get(id, 0)
                    wetlandCount[id] = wetlandPix + 1
                    wetlandID[row, col] = id
                    if stream_raster is not None and stream_raster[row, col] > 0 and reference_flow_direction_raster is None and reference_flow_direction_raster[row, col] > 0:
                        counter = 0
                        for n in range(8):
                            rowN = row + dY[n]
                            colN = col + dX[n]
                            if stream_raster[rowN, colN] > 0 and reference_flow_direction_raster[rowN, colN] == inflowingVals[n]:
                                counter += 1
                        if counter < 1 and id not in riparianWetlandRemoveIDs:
                            riparianWetlandRemoveIDs.add(id)
                        if id not in riparianWetlandIDs and id not in riparianWetlandRemoveIDs:
                            riparianWetlandIDs.add(id)

                    if id in conflictWetlands and confWetIsAdded[id] < 0:
                        grpNum = 0
                        while grpNum in confWetGroups:
                            grpNum += 1
                        confWetGroups[grpNum].append(id)
                        confWetIsAdded[id] = grpNum

                        for n in range(8):
                            rowN = row + dY[n]
                            colN = col + dX[n]
                            neighbourID = int(structure_raster[rowN, colN])
                            if neighbourID > 0 and neighbourID != id and neighbourID in conflictWetlands:
                                for grp in confWetGroups.values():
                                    if id in grp and neighbourID not in grp:
                                        grp.append(neighbourID)
                                        confWetIsAdded[neighbourID] = confWetIsAdded[id]

        biggestWet = set()
        for grpNum in confWetGroups:
            if len(confWetGroups[grpNum]) == 1:
                biggestWet.add(confWetGroups[grpNum][0])
                continue

            pix = 0
            test = 0
            for id in confWetGroups[grpNum]:
                try:
                    if wetlandCount.get(id, 0) > pix:
                        pix = wetlandCount[id]
                        if id in biggestWet:
                            biggestWet.remove(id)
                        test = id
                except Exception as e:
                    print(id)
            if test > 0 and test not in biggestWet:
                biggestWet.add(test)

        for big in biggestWet:
            conflictWetlands.remove(big)

        for removeID in riparianWetlandRemoveIDs:
            if removeID in riparianWetlandIDs:
                riparianWetlandIDs.remove(removeID)

        pixelArea = dem_raster.configs.resolution_x * dem_raster.configs.resolution_y
        isoNoOutWetlandID = set()
        isoMultiOutWetlandID = set()
        iso1OutWetlandID = set()
        for id in wetlandCount:
            area = wetlandCount[id] * pixelArea
            if area < lowerThresholdArea or id in conflictWetlands:
                isoNoOutWetlandID.add(id)

            if id in riparianWetlandIDs:
                if id in isoNoOutWetlandID:
                    riparianWetlandIDs.remove(id)
                continue

            if area > upperThresholdArea:
                isoMultiOutWetlandID.add(id)
            elif area >= lowerThresholdArea:
                iso1OutWetlandID.add(id)

        
        if len(isoMultiOutWetlandID)>0:
            structure_with_multiple_outlets_raster = self.wbe.new_raster(dem_raster.configs)

            for row in range(rows):
                for col in range(cols):
                    if wetlandID[row, col] > 0 and wetlandID[row, col] in isoMultiOutWetlandID:
                        structure_with_multiple_outlets_raster[row, col] = int(wetlandID[row, col])
                    else:
                        structure_with_multiple_outlets_raster[row, col] = nodataWet
        else:
            logger.info("No wetland bigger than threshold!")


        
        if stream_raster is not None and len(riparianWetlandIDs) > 0:
            structure_riparian_raster = self.wbe.new_raster(dem_raster.configs)

            for row in range(rows):
                for col in range(cols):
                    if wetlandID[row, col] > 0 and wetlandID[row, col] in riparianWetlandIDs:
                        structure_riparian_raster[row, col] = int(wetlandID[row, col])
                    else:
                        structure_riparian_raster[row, col] =  nodataWet


        if structure_outlet_raster is None:
            logger.info("No end node file provided!")

        for row in range(rows):
            for col in range(cols):
                if wetlandID[row, col] > 0:
                    id = int(wetlandID[row, col])
                    if id in isoNoOutWetlandID:
                        wetlandID[row, col] = -32768
                    if structure_outlet_raster is None and id in isoMultiOutWetlandID:
                        wetlandID[row, col] = -32768

                    if id in iso1OutWetlandID:
                        structure_kept_raster[row, col] = id

                    if id not in isoNoOutWetlandID:
                        structure_removed_raster[row, col] = id

                    if id in iso1OutWetlandID or id in isoMultiOutWetlandID:
                        structure_isolated_raster[row, col] = id
                else:
                    structure_kept_raster[row, col] = nodataWet

        self.save_raster(structure_kept_raster, f"{self.structure_type}Kept{Names.raster_extension}")     
        #self.save_raster(structure_removed_raster, f"{self.structure_type}BoarderLine{Names.raster_extension}")     

        id, idN, position, positionN = 0, 0, 0, 0
        minNeighbour, maxNeighbour = 0, 0
        elevation, elevationN, minElevation, maxElevation = 0.0, 0.0, 0.0, 0.0

        #flag the outlet cell for each structure
        for row in range(rows):
            for col in range(cols):
                id = wetlandID[row, col]
                if id > 0:
                    #use outlet raster layer if it's available. This assumes the outlet is at the boarder of structure
                    #if not, find the outlet along the boarder
                    if structure_outlet_raster is not None:
                        if structure_outlet_raster[row, col] > 0:
                            wetlandEndNodes[row, col] = True
                    else:
                        position = wetlandPosition[row, col]
                        minNeighbour = position
                        maxNeighbour = position
                        for n in range(8):
                            rowN = row + dY[n]
                            colN = col + dX[n]
                            positionN = wetlandPosition[rowN, colN]
                            idN = wetlandID[rowN, colN]
                            if idN == id:
                                if positionN < minNeighbour:
                                    minNeighbour = positionN
                                if positionN > maxNeighbour:
                                    maxNeighbour = positionN
                        if minNeighbour == position or maxNeighbour == position:
                            wetlandEndNodes[row, col] = True

        #find cells that is pit and edge
        output = np.full((rows, cols), nodata, dtype=np.float32)
        flowdir = np.zeros((rows, cols), dtype=np.int8)
        queue = []
        inQueue = np.zeros((rows, cols), dtype=bool)

        for row in range(rows):
            for col in range(cols):
                z = dem_raster[row, col]
                output[row, col] = z
                flowdir[row, col] = 0
                if z != nodata:
                    isWetland = wetlandID[row, col] > 0
                    isPit = True
                    isEdgeCell = False
                    for n in range(8):
                        zN = dem_raster[row + dY[n], col + dX[n]]
                        if isPit and zN != nodata and zN < z:
                            isPit = False
                        elif zN == nodata:
                            isEdgeCell = True
                    if (isPit and isEdgeCell) or (isWetland and isEdgeCell):
                        heapq.heappush(queue,GridCell(row=row, col=col, z=z, flatIndex = 0, streamVal = (not isWetland)))
                        inQueue[row, col] = True

        pit_edge_wetland_flag_raster = self.wbe.new_raster(dem_raster.configs)
        for row in range(rows):
            for col in range(cols):
                pit_edge_wetland_flag_raster[row, col] = 1 if inQueue[row, col] else nodata
        self.save_raster(pit_edge_wetland_flag_raster, f"{self.structure_type}PitEdgeStructureFlag{Names.raster_extension}")

        flatIndex = 0
        k = 0
        linkIDValue, linkIDValueN = 0, 0
        while queue:
            gc = heapq.heappop(queue)
            #logger.info(f"Row: {gc.row}, Col: {gc.col}, streamVal: {gc.streamVal}, isWetland: {not gc.streamVal} z: {gc.z}, flatIndex: {gc.flatIndex}")
            row = gc.row
            col = gc.col
            flatIndex = gc.flatIndex
            linkIDValue = wetlandID[row, col]
            isWetland = linkIDValue > 0
            if linkIDValue > 0:
                # /* move upstream following the path of 
                #  *  minimum change in link position. Stop
                #  *  when the link id changes.
                #  */
                flag = True
                r = row
                j = col
                while flag:
                    indexOfNextCell = -1
                    minPosDiff = float('inf')
                    posDiff = 0
                    position = wetlandPosition[r, j]
                    for n in range(8):
                        rowN = r + dY[n]
                        colN = j + dX[n]
                        linkIDValueN = wetlandID[rowN, colN]                        
                        #if the neibour is on the boarder
                        if linkIDValueN == linkIDValue and not inQueue[rowN, colN] and wetlandPosition[rowN, colN] > 0:
                            positionN = wetlandPosition[rowN, colN]
                            posDiff = (positionN - position) * (positionN - position)
                            if posDiff < minPosDiff:
                                minPosDiff = posDiff
                                indexOfNextCell = n
                        elif linkIDValueN == -32768 or wetlandEndNodes[rowN, colN] or wetlandPosition[rowN, colN] < 0:
                            zN = output[rowN, colN]
                            if zN != nodata and not inQueue[rowN, colN]:
                                # it's a non-stream cell or a link end node and can be added to the queue
                                flowdir[rowN, colN] = backLink[n]
                                #logger.info(f"Row = {rowN}, Col = {colN}, n = {n}, flow dir = {flowdir[rowN, colN]}")
                                k = 0
                                if zN == output[row, col]:
                                    k = flatIndex + 1
                                heapq.heappush(queue,GridCell(row=rowN, col=colN, z=zN, flatIndex = k, streamVal = (not wetlandEndNodes[rowN, colN])))
                                inQueue[rowN, colN] = True
                    if indexOfNextCell > -1:
                        rowN = r + dY[indexOfNextCell]
                        colN = j + dX[indexOfNextCell]
                        flowdir[rowN, colN] = backLink[indexOfNextCell]
                        #logger.info(f"Row = {rowN}, Col = {colN}, n = {indexOfNextCell}, flow dir = {flowdir[rowN, colN]}")
                        inQueue[rowN, colN] = True
                        r = rowN
                        j = colN
                    else:
                        #there were no unvisited neighbours of the same link ID
                        flag = False
            else:
                for n in range(8):
                    rowN = row + dY[n]
                    colN = col + dX[n]
                    if rowN >= rows or colN >= cols or rowN < 0 or colN < 0:
                        continue
                    zN = output[rowN, colN]
                    if zN != nodata and not inQueue[rowN, colN]:
                        linkIDValueN = wetlandID[rowN, colN]
                        if linkIDValueN == -32768 or wetlandEndNodes[rowN, colN]:
                            # else it's a stream and not an end node and shouldn't be added to the queue.
                            flowdir[rowN, colN] = backLink[n]
                            #logger.info(f"Row = {rowN}, Col = {colN}, n = {n}, flow dir = {flowdir[rowN, colN]}")
                            k = 0
                            if zN == output[row, col]:
                                k = flatIndex + 1
                            heapq.heappush(queue,GridCell(row=rowN, col=colN, z=zN, flatIndex = k, streamVal = (not wetlandEndNodes[rowN, colN])))
                            inQueue[rowN, colN] = True
                    

        for row in range(rows):
            for col in range(cols):
                z = output[row, col]
                if z != nodata:
                    flow_direction_raster[row, col] = outPointer[flowdir[row, col]]

        return flow_direction_raster

#endregion
 

#region Offset Outlet

    def __offset_outlets(self)->Raster:
        """
        Snap the given outlets to the structure if necessary. 
        It assumes the outlet has the same id as the structure, which is very important.

        Translate from plugin: WetlandOutletsOffset
        """

        logger.info(f"Offseting user-defined {self.structure_type} outlets ...")

        structure_raster = self.boundary_original_raster
        structure_outlet_raster = self.outlet_original_raster
        dem_raster = self.__dem_raster

        structure_outlet_offset_raster = self.wbe.new_raster(structure_outlet_raster.configs)

        rows = structure_outlet_raster.configs.rows
        cols = structure_outlet_raster.configs.columns

        wetlandMap = {}
        for row in range(rows):
            for col in range(cols):
                if structure_raster[row, col] > 0:
                    id = int(structure_raster[row, col])
                    wetlandMap.setdefault(id,[]).append((row, col))

        for row in range(rows):
            for col in range(cols):
                if structure_outlet_raster[row, col] > 0:
                    id = int(structure_outlet_raster[row, col])
                    if id in wetlandMap:
                        if id == int(structure_raster[row, col]):
                            if self.__isEdgePoint((row, col), wetlandMap[id]):
                                #if it's just at the edge, then use it directly
                                structure_outlet_offset_raster[row, col] = id
                            else:
                                #if the outlet is not at the edge, then move it to the edge
                                outlet = self.__offsetInsidePointToEdge((row, col), wetlandMap[id], dem_raster)
                                structure_outlet_offset_raster[outlet[0], outlet[1]] = id
                        else:
                            #if not in the wetland polygon, then we will need to snap to the closet point
                            outlet = self.__getClosestPoint((row, col), wetlandMap[id])
                            structure_outlet_offset_raster[outlet[0], outlet[1]] = id

        return structure_outlet_offset_raster

    def __getClosestPoint(self, checkPoint, area):
        """
        Snap the outlet to the structure
        """

        distance = float('inf')
        index = 0
        for i in range(len(area)):
            d = (area[i][0] - checkPoint[0]) ** 2 + (area[i][1] - checkPoint[1]) ** 2
            if d < distance:
                distance = d
                index = i
            if distance <= 0:
                return area[i]
        return area[index]

    def __offsetInsidePointToEdge(self, checkPoint, area, dem_raster):
        """
        Offset inside points to edge
        """
        distance = float('inf')
        indexes = []

        for i in range(len(area)):
            if self.__isEdgePoint(area[i], area):
                d = (area[i][0] - checkPoint[0]) ** 2 + (area[i][1] - checkPoint[1]) ** 2
                if d < distance:
                    distance = d
                    indexes = [i]
                elif d == distance:
                    indexes.append(i)
                if distance <= 0:
                    break

        index = indexes[0]

        if len(indexes) > 1:
            elev = float('inf')
            for i in indexes:
                elevValue = dem_raster[area[i][0], area[i][1]]
                if elevValue < elev:
                    elev = elevValue
                    index = i

        return area[index]

    def __isEdgePoint(self, checkPoint, area):
        """
        Check if the point is at edge
        """
        dX = [1, 1, 1, 0, -1, -1, -1, 0]
        dY = [-1, 0, 1, 1, 1, 0, -1, -1]
        flag = False

        for c in range(8):
            x = checkPoint[1] + dX[c]
            y = checkPoint[0] + dY[c]
            flag = not (y, x) in area and (checkPoint[0], checkPoint[1]) in area
            if flag:
                break

        return flag


#endregion

@dataclass(order=True)
class GridCell:
    row: int = field(compare=False)
    col: int = field(compare=False)
    streamVal: bool 
    z: float
    flatIndex: int


import os

class FileOut:
    def __init__(self, folder) -> None:
        self.folder = folder

    def write_file(self):
        """Use empty output file for now"""
        with open(os.path.join(self.folder, "file.out"),'w') as f:
            f.writelines("OUTPUTTYPE | 1")


from .bmp import BMP
from whitebox_workflows import Vector, Raster
import pandas as pd
from ..delineation.structure import Structure
from ..database.bmp.bmp_41_wascob import Wascob

class StructureBMPWascob(BMP):
    def __init__(self, bmp_vector:Vector, subbasin_raster:Raster, structure:Structure):
        super.__init__(bmp_vector, subbasin_raster)

        self.structure = structure

    @property
    def wascob_df(self)->pd.DataFrame:
        wascobs = [Wascob(att) for att in self.structure.attributes]
        return pd.DataFrame([vars(rb) for rb in wascobs])




import os
import math
import numpy as np
import pandas as pd
from .interpolation import Interpolation
from whitebox_workflows import Raster
from ..raster_extension import RasterExtension

class GridInterpolation(Interpolation):
    def __init__(self, weight_file: str) -> None:
        super().__init__(weight_file) 

    def write_weight_file(self, mask_raster:Raster, station_coordinates:list):
        rows = mask_raster.configs.rows
        cols = mask_raster.configs.columns
        numShapes = len(station_coordinates)
        rowCount = RasterExtension.get_number_of_valid_cell(mask_raster)

        if numShapes <= 2:
            raise ValueError("Grid Interpolation need at least three stations. ")

        sb = []
        sep = os.linesep
        sb.append(str(rowCount))
        sb.append(sep)
        sb.append(str(numShapes))
        sb.append(sep)

        xMax = mask_raster.get_x_from_column(0)
        xMin = xMax
        yMax = mask_raster.get_y_from_row(0)
        yMin = yMax

        for row in range(rows):
            for col in range(cols):
                if mask_raster[row, col] > 0:
                    xMax = max(xMax, mask_raster[row, col])
                    xMin = min(xMin, mask_raster[row, col])
                    yMax = max(yMax, mask_raster[row, col])
                    yMin = min(yMin, mask_raster[row, col])

        dist = np.zeros(numShapes)
        minDist = np.zeros(5)
        minDistStationNum = np.zeros(5, dtype=int)
        quadrant = np.zeros(numShapes, dtype=int)
        zeroes = 0
        count = np.zeros(5, dtype=int)
        finalValue = np.empty((rows, cols), dtype=object)

        for i in range(5):
            count[i] = 0
            minDist[i] = 0

        for i in range(numShapes):
            x = station_coordinates[i][0]
            y = station_coordinates[i][1]
        
            if x > xMax and y > yMax:
                quadrant[i] = 1
                count[1] += 1
            elif x > xMax and y < yMin:
                quadrant[i] = 2
                count[2] += 1
            elif x < xMin and y < yMin:
                quadrant[i] = 3
                count[3] += 1
            elif x < xMin and y > yMax:
                quadrant[i] = 4
                count[4] += 1
            else:
                quadrant[i] = 4
                count[0] += 1

        zeroes = sum(1 for i in range(1, 5) if count[i] == 0)

        if zeroes > 1:
            raise ValueError("Grid interpolation not possible!")

        for row in range(rows):
            for col in range(cols):
                if mask_raster[row, col] > 0:
                    raster_x = mask_raster.get_x_from_column(col)
                    raster_y = mask_raster.get_y_from_row(row)
                
                    for i in range(numShapes):
                        x = station_coordinates[i][0]
                        y = station_coordinates[i][1]
                        dist[i] = math.sqrt((x - raster_x) ** 2 + (y - raster_y) ** 2)

                    for i in range(numShapes):
                        for k in range(1, 5):
                            if count[k] != 0:
                                if quadrant[i] == k:
                                    if minDist[k] == 0:
                                        minDist[k] = dist[i]
                                    else:
                                        if minDist[k] > dist[i]:
                                            minDist[k] = dist[i]
                                            minDistStationNum[k] = i

                    radius = np.ones(5)
                    multiplier = np.zeros(5, dtype=int)
                    for i in range(1, 5):
                        if count[i] == 0:
                            radius[i] = 1
                            multiplier[i] = 0
                        else:
                            radius[i] = minDist[i]
                            multiplier[i] = 1

                    cValue = (multiplier[4] * (radius[1] * radius[2] * radius[3]) ** 2 +
                                multiplier[1] * (radius[2] * radius[3] * radius[4]) ** 2 +
                                multiplier[3] * (radius[1] * radius[2] * radius[4]) ** 2 +
                                multiplier[2] * (radius[1] * radius[3] * radius[4]) ** 2)

                    cValues = np.zeros(5)
                    cValuesInt = np.zeros(5, dtype=int)
                    cValues[1] = multiplier[1] * ((radius[2] * radius[3] * radius[4]) ** 2) / cValue
                    cValues[2] = multiplier[2] * ((radius[1] * radius[3] * radius[4]) ** 2) / cValue
                    cValues[3] = multiplier[3] * ((radius[1] * radius[2] * radius[4]) ** 2) / cValue
                    cValues[4] = multiplier[4] * ((radius[1] * radius[2] * radius[3]) ** 2) / cValue

                    for i in range(1, 5):
                        if count[i] == 0:
                            cValues[i] = 0
                        cValuesInt[i] = int(cValues[i] * 10000)

                    cValueIntSum = cValuesInt[1] + cValuesInt[2] + cValuesInt[3] + cValuesInt[4]
                    tempNums = np.zeros(5)

                    if cValueIntSum > 10000:
                        for i in range(1, 5):
                            tempNums[i] = cValues[i] * 10000 - math.floor(cValues[i] * 10000)
                            if tempNums[i] < 0.5:
                                tempNums[i] = 1

                        for i in range(1, 5):
                            if tempNums[i] != 0:
                                if tempNums[i] <= tempNums[1] and tempNums[i] <= tempNums[2] and tempNums[i] <= tempNums[3] and tempNums[i] <= tempNums[4]:
                                    cValuesInt[i] -= 1
                    elif cValueIntSum < 10000:
                        for i in range(1, 5):
                            tempNums[i] = cValues[i] * 10000 - math.floor(cValues[i] * 10000)
                            if tempNums[i] > 0.5:
                                tempNums[i] = 0

                        for i in range(1, 5):
                            if tempNums[i] != 0:
                                if tempNums[i] >= tempNums[1] and tempNums[i] >= tempNums[2] and tempNums[i] >= tempNums[3] and tempNums[i] >= tempNums[4]:
                                    cValuesInt[i] += 1

                    stationsNumberSst = np.zeros(5, dtype=int)
                    for i in range(1, 5):
                        stationsNumberSst[i] = minDistStationNum[i]

                    finalValues = np.zeros(numShapes, dtype=int)
                    for i in range(1, 5):
                        finalValues[stationsNumberSst[i]] = round(cValuesInt[i] / 10000)

                    finalValue[row, col] = str(finalValues[0])
                    for i in range(1, numShapes):
                        finalValue[row, col] += "\t" + str(finalValues[i])

                    sb.append(str(finalValue[row, col]))
                    sb.append(sep)

        with open(self.weight_file, 'w') as out:
            out.write(''.join(sb))






from sqlalchemy import Column, Integer, Float, String
from .bmp_table import BMPTable
from ...names import Names

class SubbasinInfo(BMPTable):
    __tablename__ = Names.bmp_table_name_subbasin_info
    ID = Column(Integer, primary_key=True)
    Area_Ha = Column(Float)
    FlowAcc_Max = Column(Float)
    PRC_Avg = Column(Float)
    DSC_Avg = Column(Float)
    CN2_Avg = Column(Float)

    def __init__(self, id, area_ha, flow_acc, prc, dsc, cn2):
        self.ID = id
        self.Area_Ha = area_ha
        self.FlowAcc_Max = flow_acc
        self.PRC_Avg = prc
        self.DSC_Avg = dsc
        self.CN2_Avg = cn2


from .bmp_management_base import BMPManagementBase

class ManureApplicationSetbackManagement(BMPManagementBase):
    """Distribution Table for BMP: Manure application setback (22)"""
    def __init__(self):
        super().__init__()



from .bmp_management_base import BMPManagementBase

class ManureApplicationBasedOnPhosphorousLimitManagement(BMPManagementBase):
    """Distribution Table for BMP: Application based on soil Phosphorous limit (25)"""
    def __init__(self):
        super().__init__()
        self.Soil_P_Limit_kg_ha = 0



from sqlalchemy import Column, Integer, Float
from .bmp_table import BMPTable
from ...names import Names

class FieldFarm(BMPTable):
    __tablename__ = Names.bmp_table_name_field_farm
    ID = Column(Integer, primary_key=True)
    Field = Column(Integer)
    Farm = Column(Integer)
    Area_Ha = Column(Float)
    ToField = Column(Float)
    ToFarm = Column(Float)


from configparser import ConfigParser
import logging

logger = logging.getLogger(__name__)

class Config:
    """Base class for all configs"""
    def __init__(self, config_file:str = None):
        self.config_file = config_file  
  
    @property
    def config_variables(self)->str:
        return {}

    def create_template(self, file_path:str):
        """
        create a template configuration file with empty values. this file function as the starting point of a working file
        """
        cf = ConfigParser()
        for section, variables in self.config_variables.items():
            cf[section] = {}
            for var in variables:
                cf[section][var] = ""
        with open(file_path, 'w') as configfile:
            cf.write(configfile)

    @staticmethod
    def get_option_value_exactly(cf:ConfigParser, section_name, option_names, valtyp=str):
        """Read the value at the given location as given type"""
        if valtyp == int:
            return cf.getint(section_name, option_names)
        elif valtyp == float:
            return cf.getfloat(section_name, option_names)
        elif valtyp == bool:
            return cf.getboolean(section_name, option_names)
        else:
            return cf.get(section_name, option_names)

    @staticmethod
    def check_config_option(cf, section_name, option_names, print_warn=False):
        """
        check if the config option exist in the config file
        """
        if not isinstance(cf, ConfigParser):
            raise IOError('ErrorInput: The first argument cf MUST be the object of `ConfigParser`!')
        if type(option_names) is not list:
            option_names = [option_names]  

        if section_name not in cf.sections():
            if print_warn:
                logger.warning(f'Warning: Section {section_name} is NOT defined, try to find in DEFAULT section!')
            for optname in option_names:  
                if cf.has_option('', optname): 
                    return True, '', optname
            if print_warn:
                logger.warning(f'Warning: Section {section_name} is NOT defined, Option {option_names} is NOT FOUND!')
            return False, '', ''
        else:
            for optname in option_names:  # For backward compatibility
                if cf.has_option(section_name, optname):
                    return True, section_name, optname
            if print_warn:
                logger.warning(f'Warning: Option {option_names} is NOT FOUND in Section {section_name}!')
            return False, '', ''

    @staticmethod
    def get_option_value(cf, section_name, option_names, valtyp=str, print_warn=True):  
        found, sname, oname = Config.check_config_option(cf, section_name, option_names, print_warn=print_warn)
        if not found:
            return None
        return Config.get_option_value_exactly(cf, sname, oname, valtyp=valtyp)


from .bmp_16_grazing_management import GrazingManagement
from typing import Any
from sqlalchemy import Column, Integer, TEXT, REAL
from .bmp_table import BMPTable
from ...names import Names

class PastureGrazingManagement(GrazingManagement):
    """Distribution Table for BMP: Pasture grazing management (37)"""
    def __init__(self):
        super().__init__()


class PastureGrazingParameter(BMPTable):
    """Distribution Table for BMP: Pasture grazing management (37)"""
    __tablename__ = Names.bmp_table_name_pasture_grazing_parameter
    """Pasture grazing land ID"""
    ID = Column(Integer, primary_key=True)
    """Producer ID"""
    ProducerID = Column(Integer)
    """Pasture grazing area (ha)"""
    Area_ha = Column(REAL)


from .bmp import BMP
from whitebox_workflows import Vector, Raster
from ..raster_extension import RasterExtension
from typing import Dict, List, Tuple
import math
import numpy as np

class ArealBMP(BMP):
    def __init__(self,
                 bmp_vector:Vector,
                 flow_direction_raster:Raster,
                 subbasin_raster:Raster,
                 reach_raster:Raster,
                 dem_raster:Raster):
        super().__init__(bmp_vector, subbasin_raster)

        self.flow_direction_raster = flow_direction_raster
        self.reach_raster = reach_raster
        self.dem_raster = dem_raster
        
        self.__contribution_area_ha = None
        self.__contribution_area_subbasin_ratio = None
        self.__bmp_locations = None
        self.__reach_ids = None
        self.__reach_distance = None
        self.__reach_elevation_drop = None    
   
    @property
    def contribution_area_ha(self)->dict:
        if self.__contribution_area_ha is None:
            self.__contribution_area_ha = RasterExtension.get_zonal_statistics(self.flow_direction_raster, self.bmp_raster,"max","contribution_area_ha") * self.subbasin_raster.configs.resolution_x * self.subbasin_raster.configs.resolution_y / 10000.0 
            self.__contribution_area_ha = self.__contribution_area_ha["contribution_area_ha"].to_dict()

        return self.__contribution_area_ha
    
    @property
    def contribution_area_subbasin_ratio(self)->dict:
        if self.__contribution_area_subbasin_ratio is None:
            self.__contribution_area_subbasin_ratio = {}

            subbasin_area_ha = RasterExtension.get_category_area_ha_dataframe(self.subbasin_raster, "subbasin_area_ha")["subbasin_area_ha"].to_dict()
            for id, subbasin in self.subbasins:
                self.__contribution_area_subbasin_ratio[id] = self.contribution_area_ha[id] / subbasin_area_ha[subbasin]

        return self.__contribution_area_subbasin_ratio

    @property
    def bmp_locations(self) -> Dict[int, List[Tuple[int, int]]]:
        """The row and col index of each bmp"""
        if self.__bmp_locations is None:
            self.__bmp_locations = {}        
            rows = self.bmp_raster.configs.rows
            cols = self.bmp_raster.configs.columns
            for row in range(rows):
                for col in range(cols):
                    value = self.bmp_raster[row, col]
                    if value > 0:
                        id = int(value)
                        if id not in self.__bmp_raster:
                            self.__bmp_raster[id] = []
                        self.__bmp_raster[id].append((row, col))
            
        return self.__bmp_locations
        
    @property
    def reach_ids(self)->dict:
        """The downstream reach id for each bmp"""
        if self.__reach_ids is None:
            self.__reach_ids = {}
            for id, cells in self.bmp_locations.items():
                for cell in cells:
                    reach = self.__get_down_stream_row_col(cell[0], cell[1], -1)
                    if reach is not None:
                        self.__reach_ids[id] = self.reach_raster[reach[0], reach[1]]
                        break

                if id not in self.__reach_ids:
                    self.__reach_ids[id] = -1

        return self.__reach_ids
    
    @property
    def reach_distance(self)->dict:
        """the distance to reach"""
        if self.__reach_distance is None: 
            self.__reach_distance = {}
            for id, cells in self.bmp_locations.items():
                out = -1
                for cell in cells:
                    distance = self.__get_downstream_distance(cell[0], cell[1], self.reach_ids[id])
                    if distance > 0 and (distance < out or out < 0):
                        out = distance

                self.__reach_distance[id] = max(out, 0)

        return self.__reach_distance
    
    @property 
    def reach_elevation_drop(self)->dict:
        """The elevation drop from the bmp to the reach"""
        if self.__reach_elevation_drop is None:
            self.__reach_elevation_drop = {}
            for id, cells in self.bmp_locations.items():
                out = -1
                for cell in cells:
                    reach = self.__get_down_stream_row_col(cell[0], cell[1], self.reach_ids[id])
                    if reach is not None:
                        elevation_drop = self.dem_raster[cell[0], cell[1]] - self.dem_raster[reach[0], reach[1]]
                        if elevation_drop > 0 and (elevation_drop < out or out < 0):
                            out = elevation_drop

                self.__reach_elevation_drop[id] = max(out, 0)     

        return self.__reach_elevation_drop      
    
    def __get_down_stream_row_col(self, row, col, reach_id):
        """get first row and col of downstream reach"""
        dx = [1, 1, 1, 0, -1, -1, -1, 0]
        dy = [-1, 0, 1, 1, 1, 0, -1, -1]
        ln_of_2 = np.log(2)
        x, y = col, row

        while True:
            flow_dir = self.flow_direction_raster[y, x]
            if flow_dir > 0:
                c = int(np.log(flow_dir) / ln_of_2)
                if c > 7:
                    return [0, 0]
                if self.reach_raster[y, x] > 0 and reach_id == int(self.reach_raster[y, x]):
                    return [y, x]
                elif reach_id < 0 and self.reach_raster[y, x] > 0:
                    return [y, x]

                x += dx[c]
                y += dy[c]
            else:
                return None

        return None

    def __get_downstream_distance(self, row, col, reach_id):
        diag_dist = math.sqrt(self.reach_raster.configs.resolution_x ** 2 + self.reach_raster.configs.resolution_y ** 2)

        dx = [1, 1, 1, 0, -1, -1, -1, 0]
        dy = [-1, 0, 1, 1, 1, 0, -1, -1]
        dir_dist = [diag_dist, self.reach_raster.configs.resolution_x, diag_dist, self.reach_raster.configs.resolution_y,
                    diag_dist, self.reach_raster.configs.resolution_x, self.reach_raster.configs.resolution_y, diag_dist]
        ln_of_2 = math.log(2)

        x, y = col, row
        total_dist = 0

        while True:
            flow_dir = self.flow_direction_raster[y, x]
            if flow_dir > 0:
                c = int(math.log(flow_dir) / ln_of_2)
                if c > 7:
                    return -1
                if self.reach_raster[y, x] > 0 and reach_id == int(self.reach_raster[y, x]):
                    break
                elif reach_id < 0 and self.reach_raster[y, x] > 0:
                    break

                x += dx[c]
                y += dy[c]
                total_dist += dir_dist[c]
            else:
                return -1

        return total_dist



import numpy as np
import pandas as pd
from .reach_parameter import ReachParameter
from ...raster_extension import RasterExtension
from whitebox_workflows import Raster
import math

import logging
logger = logging.getLogger(__name__)

class Reach:
    """
    This genenates all the reach parameters. Some of the parameters use default values. 
    The ReachParameter class provides the table structure and default values. 
    This replaces the Reach.Java with perforamce improvements with whitebox workflow built-in functions like zonal_statistics.
    The loops should avoided whenever possible. 

    Replace buildReachParTable
    """
    def __init__(self, 
                 dem_raster:Raster, 
                 subbasin_raster:Raster, 
                 stream_network_raster:Raster, 
                 stream_order_raster:Raster, 
                 flow_length_raster:Raster, 
                 flow_direction_raster:Raster,
                 flow_accumulation_raster:Raster,
                 reach_width_raster:Raster, 
                 reach_depth_raster:Raster,                 
                 velocity_raster:Raster):
        self.LnOf2 = 0.693147180559945
        self.minimumSlope = 0.01

        self.subbasin_raster = subbasin_raster  
        self.stream_network_raster = stream_network_raster
        self.reach_length_raster = flow_length_raster
        self.flow_dir_raster = flow_direction_raster
        self.stream_order_raster = stream_order_raster
        self.reach_width_raster = reach_width_raster
        self.reach_depth_raster = reach_depth_raster
        self.dem_raster = dem_raster
        self.velocity_raster = velocity_raster
        self.flow_acc_raster = flow_accumulation_raster
        
        self.subMax = int(RasterExtension.get_max_value(self.subbasin_raster))
        self.rowNum = int(self.stream_network_raster.configs.rows)
        self.colNum = int(self.stream_network_raster.configs.columns)

        self.cellsize = dem_raster.configs.resolution_x
        self.cellsize_ha = dem_raster.configs.resolution_x * dem_raster.configs.resolution_y / 10000    
        self.min_length = (dem_raster.configs.resolution_x + dem_raster.configs.resolution_y) / 4
        self.nodata = dem_raster.configs.nodata

        self.dX = [1, 1, 1, 0, -1, -1, -1, 0]
        self.dY = [-1, 0, 1, 1, 1, 0, -1, -1]

    def __calculate_reach_length(self):
        rchmax = np.zeros(self.subMax)
        rchmin = np.zeros(self.subMax)
        rch_in_row = np.zeros(self.subMax, dtype=int)  # the row of in point of main stream
        rch_in_col = np.zeros(self.subMax, dtype=int)  # the col of in point of main stream
        rch_in_row2 = np.zeros(self.subMax, dtype=int)  # the row of in point of main stream which doesn't has upstream
        rch_in_col2 = np.zeros(self.subMax, dtype=int)  # the row of in point of main stream which doesn't has upstream
        currentSub = 0


        for i in range(0, self.subMax):
            rchmax[i] = 0
            rchmin[i] = self.reach_length_raster.configs.maximum
            rch_in_row[i] = -1
            rch_in_col[i] = -1
            rch_in_row2[i] = -1
            rch_in_col2[i] = -1

        for row in range(self.rowNum):
            for col in range(self.colNum):
                currentValue = self.reach_length_raster[row, col]
                currentSub = int(self.subbasin_raster[row, col])
                if currentValue > 0 and currentSub > 0 and self.stream_network_raster[row, col] > 0:
                    if currentValue > rchmax[currentSub - 1]:
                        rchmax[currentSub - 1] = currentValue

                    if currentValue < rchmin[currentSub - 1]:
                        rchmin[currentSub - 1] = currentValue
                        rch_in_row2[currentSub - 1] = row
                        rch_in_col2[currentSub - 1] = col

                    if rch_in_row[currentSub - 1] == -1:
                        flowDir = -1
                        previousSubID = currentSub
                        nextRow = row
                        nextCol = col
                        flag = True
                        while flag:
                            flowDir = self.flow_dir_raster[nextRow, nextCol]
                            if flowDir > 0:
                                c = int(np.log(flowDir) / self.LnOf2)
                                if c > 7:
                                    print("An unexpected value has been identified in the pointer image. This tool requires a pointer grid that has been created using either the D8 or Rho8 tools.")
                                    flag = False
                                    break
                                nextCol += self.dX[c]
                                nextRow += self.dY[c]
                                if self.stream_network_raster[nextRow, nextCol] < 0:  # not stream
                                    flag = False
                                    break
                                subID = int(self.subbasin_raster[nextRow, nextCol])
                                if subID < 0:
                                    flag = False
                                    break
                                if previousSubID != subID:  # go into next subbasin
                                    if rch_in_row[subID - 1] != -1:  # already found
                                        flag = False
                                        break
                                    rch_in_row[subID - 1] = nextRow
                                    rch_in_col[subID - 1] = nextCol
                                    previousSubID = subID

                                if self.reach_length_raster[nextRow, nextCol] >= self.reach_length_raster.configs.maximum:  # outlet
                                    flag = False
                                    break
                            else:
                                flag = False
                                break

        for i in range(self.subMax):
            if rch_in_row[i] != -1:
                rchmin[i] = self.reach_length_raster[rch_in_row[i], rch_in_col[i]]
            else:
                rchmin[i] = 0
                rch_in_row[i] = rch_in_row2[i]
                rch_in_col[i] = rch_in_col2[i]

        return rch_in_row, rch_in_col, np.maximum(rchmax - rchmin, self.min_length) 

    def __calculate_reach_mean(self, value_raster, name:str, min_value = None):
        #filter out the raster that are not on a stream       
        raster = self.stream_network_raster.con(f"value == {self.stream_network_raster.configs.nodata}", self.nodata, value_raster)

        #get the mean width in each subbasin
        reach_width_df = RasterExtension.get_zonal_statistics(raster, self.subbasin_raster, "mean", name)

        #use min value if available
        if min_value is not None:
            reach_width_df[name] = np.where(reach_width_df[name] < min_value, min_value, reach_width_df[name])

        #return the np array
        return reach_width_df[name].to_numpy()

    def __calculate_reach_width(self):
        """
        Mean width of cells on each reach        
        """
        return self.__calculate_reach_mean(self.reach_width_raster, "width", 0.1)

    def __calculate_reach_depth(self)->pd.DataFrame:
        """
        Mean depth of cells on each reach        
        """
        return self.__calculate_reach_mean(self.reach_depth_raster, "depth", 0.1)
    
    def __calculate_reach_velocity(self):
        """
        Mean velocity of cells on each reach        
        """
        return self.__calculate_reach_mean(self.velocity_raster, "veloctiy")

    def __calculate_reach_slope(self, rch_in_row, rch_in_col):
        rchavg = np.zeros(self.subMax)

        dx1 = self.cellsize
        dx2 = dx1 * np.sqrt(2.0)

        for i in range(self.subMax):
            totalSlope = 0.0
            totalCell = 1  # H.Shao >> 160404-01 <<
            flowDir = -1
            previousRow = rch_in_row[i]
            previousCol = rch_in_col[i]
            nextRow = rch_in_row[i]
            nextCol = rch_in_col[i]

            flag = True
            while flag:
                flowDir = self.flow_dir_raster[nextRow, nextCol]
                if flowDir > 0:
                    c = int(np.log(flowDir) / self.LnOf2)
                    if c > 7:
                        print("An unexpected value has been identified in the pointer image. This tool requires a pointer grid that has been created using either the D8 or Rho8 tools.")
                        flag = False
                        break
                    nextCol += self.dX[c]
                    nextRow += self.dY[c]

                    dy1 = self.dem_raster[previousRow, previousCol] - self.dem_raster[nextRow, nextCol]
                    dx3 = self.dX[c] * self.dY[c] == 0 and dx1 or dx2

                    totalSlope += dy1 / dx3
                    totalCell += 1

                    previousRow = nextRow
                    previousCol = nextCol

                    subID = int(self.subbasin_raster[nextRow, nextCol])
                    if i != subID - 1:  # go into next subbasin or border, then stop
                        flag = False
                        break
                else:
                    flag = False
                    break

            rchavg[i] = 0
            rchavg[i] = np.maximum(totalSlope / totalCell * 100.0, self.minimumSlope)

        return rchavg
    
    def __calculate_reach_order(self):
        """
        Just get max order for each subbasin
        """

        reach_order_df = RasterExtension.get_zonal_statistics(self.stream_order_raster, self.subbasin_raster, ["max"], "order")
        default_reach_order_df = pd.DataFrame(index=range(1, self.subMax + 1))
        default_reach_order_df["order"] = 1

        reach_order_df = reach_order_df.combine_first(default_reach_order_df)

        #return the np array
        return reach_order_df["order"].to_numpy()

    def __calculate_reach_manning(self,rch_in_row, rch_in_col):
        """
        Manning is linearly interpolated from a min and max manning based on stream order raster. 

        It seems there is no need to do the way it is. Just simply loop through each reach and do the 
        linear interpolation as the stream order is calculated for each reach.
        """
        rchavg = np.zeros(self.subMax)

        minReachManning = -0.1
        maxReachManning = -0.04
        unitManning = (maxReachManning - minReachManning) / (self.stream_order_raster.configs.maximum - self.stream_order_raster.configs.minimum)

        for i in range(self.subMax):
            totalManning = 0.0
            totalManningNum = 0
            nextStreamOrder = -1
            previousStreamOrder = -1
            nextRow = rch_in_row[i]
            nextCol = rch_in_col[i]

            flag = True
            while flag:
                # calculate manning
                nextStreamOrder = self.stream_order_raster[nextRow, nextCol]
                if previousStreamOrder != nextStreamOrder:
                    totalManning += minReachManning + unitManning * (nextStreamOrder - self.stream_order_raster.configs.minimum)
                    totalManningNum += 1
                    previousStreamOrder = nextStreamOrder

                flowDir = self.flow_dir_raster[nextRow, nextCol]
                if flowDir > 0:
                    c = int(np.log(flowDir) / self.LnOf2)
                    if c > 7:
                        print("An unexpected value has been identified in the pointer image. This tool requires a pointer grid that has been created using either the D8 or Rho8 tools.")
                        flag = False
                        break
                    nextCol += self.dX[c]
                    nextRow += self.dY[c]
                else:
                    flag = False
                    break

                subID = int(self.subbasin_raster[nextRow, nextCol])
                if i != subID - 1:
                    flag = False
                    break

            rchavg[i] = max(min(-1 * totalManning / totalManningNum, 0.1), 0.04)

        return rchavg

    def __calculate_receive_reach_id(self):
        rchavg = np.zeros(self.subMax, dtype=int)

        for i in range(self.subMax):
            rchavg[i] = -1

        for row in range(self.rowNum):
            for col in range(self.colNum):
                currentSub = int(self.subbasin_raster[row, col])
                if currentSub > 0:
                    for i in range(self.subMax):
                        if rchavg[i] == -1 and currentSub == i + 1:
                            currentRch = currentSub
                            nextCol = col
                            nextRow = row

                            flag = True
                            flowDir = -1
                            # H.Shao >> 160510-01
                            checkDownStream = 5  # Check 5 downstreams to see if there is loop flow path
                            checkList = []
                            tempRch = -1
                            # H.Shao << 160510-01
                            while flag:
                                flowDir = self.flow_dir_raster[nextRow, nextCol]
                                if flowDir > 0:
                                    c = int(np.log(flowDir) / self.LnOf2)
                                    if c > 7:
                                        print("An unexpected value has been identified in the pointer image. This tool requires a pointer grid that has been created using either the D8 or Rho8 tools.")
                                        flag = False
                                        break
                                    nextCol += self.dX[c]
                                    nextRow += self.dY[c]

                                    nextRch = int(self.subbasin_raster[nextRow, nextCol])

                                    if nextRch != currentRch or checkList:
                                        if not checkList:
                                            tempRch = nextRch

                                        if nextRch not in checkList:
                                            checkList.append(nextRch)

                                        if nextRch == currentRch:
                                            checkList = []
                                            tempRch = -1
                                            continue
                                        if (len(checkList) == checkDownStream and currentRch not in checkList) or nextRch < 0:
                                            if tempRch > 0:
                                                rchavg[currentRch - 1] = tempRch
                                                flag = False
                                            else:
                                                rchavg[currentRch - 1] = 0
                                                flag = False

                                else:
                                    if tempRch > 0:
                                        rchavg[currentRch - 1] = tempRch
                                        flag = False
                                    else:
                                        rchavg[currentRch - 1] = 0
                                        flag = False
                                    flag = False
                                    break

        return np.maximum(rchavg, 0)

    def __calculate_contribution_area(self):
        """
        Basically get the max flow_acc in each subbasin and multipy the cell area
        """
        #get the mean depth in each subbasin
        reach_contribution_area_df = RasterExtension.get_zonal_statistics(self.flow_acc_raster, self.subbasin_raster, "max", "contri")

        #get the area
        reach_contribution_area_df = reach_contribution_area_df * self.cellsize_ha

        #return numpy array
        #logger.info(reach_contribution_area_df)
        return reach_contribution_area_df["contri"].to_numpy()

    def __calculate_reach_elevation(self):
        #filter out the raster that are not on a stream       
        raster = self.stream_network_raster.con(f"value == {self.stream_network_raster.configs.nodata}", self.nodata, self.dem_raster)

        #get the mean width in each subbasin
        reach_elevation_df = RasterExtension.get_zonal_statistics(raster, self.subbasin_raster, ["max","min","mean"])

        #return the np array
        return reach_elevation_df["max"].to_numpy(), reach_elevation_df["min"].to_numpy(),reach_elevation_df["mean"].to_numpy()

    @property
    def reach_parameter_df(self)->pd.DataFrame:
        reach_parameters = []
        for row in range(self.subMax):
            reach_parameters.append(ReachParameter())
        reach_parameter_table = pd.DataFrame([p.to_dict() for p in reach_parameters])
        reach_parameter_columns = reach_parameter_table.columns

        # Column 0 --- Reach ID "#reach_id"
        reach_parameter_table[reach_parameter_columns[0]] = range(1, self.subMax + 1)
        # Column 1 --- Subbasin ID "subbasin_id"
        reach_parameter_table[reach_parameter_columns[1]] = range(1, self.subMax + 1)
        # Column 2 --- Outlet ID "outlet_id"
        reach_parameter_table[reach_parameter_columns[2]] = range(1, self.subMax + 1)

        # Column 3 --- Reach length "length"
        rch_in_row, rch_in_col, reach_length = self.__calculate_reach_length()
        reach_parameter_table[reach_parameter_columns[3]] = reach_length

        # Column 4 --- Reach width "width"
        reach_parameter_table[reach_parameter_columns[4]] = self.__calculate_reach_width()

        # Column 8 --- Reach depth "depth"
        reach_parameter_table[reach_parameter_columns[8]] = self.__calculate_reach_depth()

        # Column 9 --- Reach slope "slope" -- unit of %
        reach_parameter_table[reach_parameter_columns[9]] = self.__calculate_reach_slope(rch_in_row, rch_in_col)

        # Column 10 --- Reach order "order"       
        reach_parameter_table[reach_parameter_columns[10]] = self.__calculate_reach_order()

        # Column 11 --- Manning's coefficient "manning"
        reach_parameter_table[reach_parameter_columns[11]] = self.__calculate_reach_manning(rch_in_row, rch_in_col)

        # Column 12 --- Velocity "velocity"
        reach_parameter_table[reach_parameter_columns[12]] = self.__calculate_reach_velocity()

        # Column 15, 16, 17 --- Max, Min, and Average elevation "max_elevation", "min_elevation", "ave_elevation"
        max_elev, min_elev, ave_elev = self.__calculate_reach_elevation()
        reach_parameter_table[reach_parameter_columns[15]] = max_elev
        reach_parameter_table[reach_parameter_columns[16]] = min_elev
        reach_parameter_table[reach_parameter_columns[17]] = ave_elev

        # Column 18 --- Receiving stream reach ID "receive_reach_id"
        reach_parameter_table[reach_parameter_columns[18]] = self.__calculate_receive_reach_id()

        # Column 19 --- Subbasin contribution area "contribution_area"
        reach_parameter_table[reach_parameter_columns[19]] = self.__calculate_contribution_area() 

        return reach_parameter_table





from typing import Any
from sqlalchemy import Column, TEXT, REAL, INTEGER
from .bmp_table import BMPTable
from ...names import Names


class WaterUse(BMPTable):
    __tablename__ = Names.bmp_table_name_point_source
    ID = Column(INTEGER, primary_key=True)
    SourceType = Column(INTEGER)
    SourceID = Column(INTEGER)
    OPERATION = Column(TEXT)
    WaterUseDataTable = Column(TEXT)

    def __init__(self):
        self.XLL = 0
        self.YLL = 0
        self.OPERATION = "1900-01-01"
        self.TABLENAME = ""


from .bmp_management_base import BMPManagementBase

class ManureApplicationBasedOnSoliNitrogenLimitManagement(BMPManagementBase):
    """Distribution Table for BMP: Application based on soil nitrogen limit (26)"""
    def __init__(self):
        super().__init__()
        self.NO3_N_Limit_kg_ha = 0



from configparser import ConfigParser
import os
from whitebox_workflows import WbEnvironment, Vector
from ..raster_extension import RasterExtension
from ..vector_extension import VectorExtension
from ..names import Names
import shutil
import logging
from .config import Config
from pathlib import Path
from ..model.model import Model

logger = logging.getLogger(__name__)

class ModelConfig(Config):
    """
    Config for watershed delineation and model structure creation
    """

    #the sections that are files
    file_sections = ["watershed","lookup","database","reach_bmp","structure_bmp","non_structure_bmp"]

    def __init__(self, config_file:str = None):
        super().__init__(config_file)
        self.input_folder = ""
        self.model = None
        
        self.__load()

    def __load(self):
        """
        load the config file, validate and copy the input files
        """

        if self.config_file is not None and os.path.exists(self.config_file):
            rasters = {}
            vectors = {}
            unique_id_vectors = {}
            databases = {}
            lookups = {}

            wbe = WbEnvironment()
            
            logger.info(f"Loading model configuration file {self.config_file} ...")
            cf = ConfigParser()
            cf.read(self.config_file)
            for section, variables in self.config_variables.items():
                for var in variables:
                    #logger.info(f"Reading Section: {section}, Variable: {var} ...")
                    if section == "delineation":
                        value = Config.get_option_value_exactly(cf, section, var, valtyp=float)
                        setattr(self, var, value)
                        continue
                    else:
                        value = Config.get_option_value(cf, section, var)
                        setattr(self, var, value)

                    if value is None or len(value) <= 0:
                        continue

                    #check input folder
                    if var == "input_folder":
                        if os.path.exists(value):
                            self.input_folder = value
                        else:
                            raise ValueError(f"{var} = {value} doesn't exist!")
                    
                    #create model folder
                    if var == "model_folder":
                        if not os.path.exists(value):
                            os.makedirs(value)
                        self.model = Model(value)
                        continue

                    if section in ModelConfig.file_sections:
                        file_path = os.path.join(self.input_folder, value)
                        if os.path.exists(file_path):
                            if "raster" in var:
                                rasters[var] = wbe.read_raster(file_path)
                            if "shapefile" in var:
                                vectors[var] = wbe.read_vector(file_path)
                                if "bmp" in section:
                                    unique_id_vectors[var] = vectors[var]
                            if section == "database":
                                databases[var] = value
                            if section == "lookup":
                                lookups[var] = value
                        else:
                            raise ValueError(f"{var} = {value} doesn't exist!")

            #check rasters and vectors
            RasterExtension.check_rasters(rasters)
            VectorExtension.check_vectors(unique_id_vectors)

            #copy all the input files to the model folder with standard name
            self.model.create_model_folder()
            logger.info(f"Copying input files to {self.input_folder} ...")

            logger.info("Rasters ...")
            for option, raster in rasters.items():
                logger.info(raster.file_name)
                wbe.write_raster(raster, os.path.join(self.model.model_input_folder, Names.config_item_standard_name_lookup[option]))
            
            logger.info("Vectors ...")
            for option, vector in vectors.items():
                logger.info(vector.file_name)
                VectorExtension.save_vector(vector, os.path.join(self.model.model_input_folder, Names.config_item_standard_name_lookup[option]))
            
            logger.info("Databases ...")
            for option, databae_file in databases.items():
                logger.info(databae_file)
                shutil.copyfile(os.path.join(self.input_folder, databae_file), os.path.join(self.model.model_database_folder, Names.config_item_standard_name_lookup[option]))
            
            logger.info("Lookup files ...")
            for option, lookup_file in lookups.items():
                logger.info(lookup_file)
                source_file = os.path.join(self.input_folder, lookup_file)
                target_file = os.path.join(self.model.model_input_folder, Names.config_item_standard_name_lookup[option])

                #we need to write the lookup file to utf8 so it could be loaded correctly in pandas later.
                Path(target_file).write_text(Path(source_file).read_text(), encoding="utf8") 

    @property
    def config_variables(self)->str:
        return {
            "default":["input_folder"],

            #watershed definition
            "watershed":["dem_raster",
                         "soil_raster",
                         "landuse_raster",
                         "stream_shapefile",
                         "boundary_shapefile",
                         "farm_shapefile",
                         "field_shapefile",
                         "outlet_shapefile" #not used right now
                        ],

            #lookup tables            
            "lookup":["soil_lookup",
                      "landuse_lookup"],

            #db3 databases
            "database":["hydroclimate"],

            #we only need the shapefile of bmps that will impact the watershed dilineation.
            #all existing and future ones should be included in the shapefile and could be
            #enabled or disabled in scenario design stage.
            #also assume that all shapefile has an ID column as the structure ids. This will be enforced. 
            "reach_bmp":[ "point_source_shapefile",
                          "flow_diversion_shapefile",
                          "reservoir_shapefile",
                          "wetland_boundary_shapefile",
                          "wetland_outlet_shapefile",
                          "manure_catch_basin_shapefile",
                          "grass_waterway_shapefile",
                          "access_management_shapefile",
                          "water_use_shapefile"
                          ],

            #wasco will be the outlet of the subbasin
            "structure_bmp":["dugout_boundary_shapefile",
                             "wascob_boundary_shapefile",
                             "riparian_buffer_shapefile",
                             "filter_strip_shapefile",
                             "tile_drain_shapefile"],

            #feedlot will be delineated in a single subbasin, the catch basin will function as the outlet.
            "non_structure_bmp":["manure_feedlot_boundary_shapefile",
                                 "manure_feedlot_outlet_shapefile",
                                 "manure_storage_boundary_shapefile",
                                 "offsite_watering_shapefile"],

            "reservoir":["reservoir_flow_routing","reservoir_flow_data_folder"],

            #delineation parameters
            "delineation":["stream_threshold_area_ha",
                        "wetland_min_area_ha",
                        "wetland_riparian_contribution_area_ha",
                        "wetland_stream_buffer_distance_m"],

            "marginal_crop_land":["non_agriculture_landuse_ids","buffer_size_m","slope_threshold_percentage"],

            "pasture_land":["non_agriculture_landuse_ids"],

            #crop rotation
            "crop_rotation":["method",
                              "AAFC_crop_inventory_folder",
                              "first_year",
                              "last_year"],

            #model folder
            "model":["model_folder"]
        }
    
    def delineate_watershed(self):
        """watershed delineation""" 
        self.model.delineate_watershed()

    def generate_parameters(self):
        """watershed delineation""" 
        self.model.generate_parameters(getattr(self,"reservoir_flow_routing"), 
                                       getattr(self,"reservoir_flow_data_folder"))

    def update_crop_rotation(self):
        self.model.update_crop_rotation(
            getattr(self,"AAFC_crop_inventory_folder"),
            int(getattr(self,"first_year")),
            int(getattr(self,"last_year"))
        )


import os
from .interpolation import Interpolation
from whitebox_workflows import Raster, WbEnvironment
import numpy as np
from ..raster_extension import RasterExtension

class LinearTriangle(Interpolation):
    def __init__(self, weight_file: str) -> None:
        super().__init__(weight_file) 

    def write_weight_file(self, mask_raster:Raster, station_coordinates:list): 
        rows = mask_raster.configs.rows
        cols = mask_raster.configs.columns
        numShapes = len(station_coordinates)
        rowCount = RasterExtension.get_number_of_valid_cell(mask_raster)

        sb = []
        sep = os.linesep
        sb.append(str(rowCount))
        sb.append(sep)
        sb.append(str(numShapes))
        sb.append(sep)

        if numShapes <= 2:
            raise ValueError("Linear Triangle need at least three stations. ")

        dist = np.zeros(numShapes)
        d123 = np.zeros(3)
        stationsNum = np.zeros(3, dtype=int)
        aValues = np.zeros(3)
        aValuesInt = np.zeros(3, dtype=int)
        x = np.zeros(3)
        y = np.zeros(3)

        finalValue = [[""] * cols for _ in range(rows)]
        for row in range(rows):
            for col in range(cols):
                if mask_raster[row, col] > 0:
                    d123[:] = np.inf
                    stationsNum[:] = 0
                    x[:] = 0
                    y[:] = 0

                    raster_x = mask_raster.get_x_from_column(col)
                    raster_y = mask_raster.get_y_from_row(row)

                    x = station_coordinates[i][0]
                    y = station_coordinates[i][1]

                    for i in range(numShapes):
                        dist[i] = np.sqrt((station_coordinates[i][0] - raster_x) ** 2 +
                                            (station_coordinates[i][1] - raster_y) ** 2)

                        if dist[i] < d123[0]:
                            d123[2] = d123[1]
                            stationsNum[2] = stationsNum[1]
                            x[2] = x[1]
                            y[2] = y[1]

                            d123[1] = d123[0]
                            stationsNum[1] = stationsNum[0]
                            x[1] = x[0]
                            y[1] = y[0]

                            d123[0] = dist[i]
                            stationsNum[0] = i
                            x[0] = x
                            y[0] = y
                        elif dist[i] < d123[1]:
                            d123[2] = d123[1]
                            stationsNum[2] = stationsNum[1]
                            x[2] = x[1]
                            y[2] = y[1]

                            d123[1] = dist[i]
                            stationsNum[1] = i
                            x[1] = x
                            y[1] = y
                        elif dist[i] < d123[2]:
                            d123[2] = dist[i]
                            stationsNum[2] = i
                            x[2] = x
                            y[2] = y

                    sValues = np.zeros(3)
                    sValuesSum = 0
                    for i in range(3):
                        sValues[0] = abs(raster_x * y[1] +
                                        (raster_y * x[2]) +
                                        (x[1] * y[2]) - (y[1] * x[2]) -
                                        (raster_x * x[1]) -
                                        (raster_y * y[2]))
                        sValues[1] = abs((x[0] * raster_y) +
                                        (y[0] * x[2]) +
                                        (raster_x * y[2]) -
                                        (raster_y * x[2]) -
                                        (y[0] * raster_x) -
                                        (x[0] * y[2]))
                        sValues[2] = abs((x[0] * y[1]) +
                                        (y[0] * raster_x) +
                                        (x[1] * raster_y) -
                                        (y[1] * raster_x) -
                                        (y[0] * x[1]) -
                                        (x[0] * raster_y))
                        sValues[0] = abs(sValues[0])
                        sValues[1] = abs(sValues[1])
                        sValues[2] = abs(sValues[2])
                        sValuesSum = sValues[0] + sValues[1] + sValues[2]

                    aValuesIntSum = 0

                    for i in range(3):
                        aValues[i] = sValues[i] / sValuesSum
                        aValuesInt[i] = int(aValues[i] * 10000)
                        aValuesIntSum += aValuesInt[i]

                    if aValuesIntSum > 10000:
                        tempNums = np.zeros(3)
                        for i in range(3):
                            tempNums[i] = aValues[i] * 10000 - np.floor(aValues[i] * 10000)
                            if tempNums[i] < 0.5:
                                tempNums[i] = 1

                        for i in range(3):
                            if tempNums[i] != 1:
                                if tempNums[i] <= tempNums[0] and tempNums[i] <= tempNums[1] and tempNums[i] <= tempNums[2]:
                                    aValuesInt[i] = aValuesInt[i] - 1
                    elif aValuesIntSum < 10000:
                        tempNums = np.zeros(3)
                        for i in range(3):
                            tempNums[i] = aValues[i] * 10000 - np.floor(aValues[i] * 10000)
                            if tempNums[i] > 0.5:
                                tempNums[i] = 0

                        for i in range(3):
                            if tempNums[i] != 1:
                                if tempNums[i] >= tempNums[0] and tempNums[i] >= tempNums[1] and tempNums[i] >= tempNums[2]:
                                    aValuesInt[i] = aValuesInt[i] + 1

                    finalValues = np.zeros(numShapes, dtype=int)

                    for i in range(numShapes):
                        finalValues[i] = 0
                    for i in range(3):
                        finalValues[stationsNum[i]] = aValuesInt[i]

                    finalValue[row][col] = str(round(float(finalValues[0]) / 10000))
                    for i in range(1, numShapes):
                        finalValue[row][col] += "\t" + str(round(float(finalValues[i]) / 10000))

                    sb.append(str(finalValue[row][col]))
                    sb.append(sep)

        with open(self.weight_file, 'w') as out:
            out.write(''.join(sb))





from typing import Any
from sqlalchemy import Column, Integer, REAL, TEXT
from .bmp_table import BMPTable
from ...names import Names

class ManureCatchBasinParameter(BMPTable):
    """Parameter Table for BMP: Manure catch basin/impondment (28)"""
    __tablename__ = Names.bmp_table_name_manure_catch_basin
    ID = Column(Integer, primary_key = True)
    OperationYear = Column(Integer)
    Subbasin = Column(Integer)
    receive_reach_id = Column(Integer)
    MaxL_m = Column(REAL)
    MaxW_m = Column(REAL)
    MaxD_m = Column(REAL)
    SideSlope = Column(REAL)
    EndSlope = Column(REAL)
    MaxVolume_104M3 = Column(REAL)
    FDLT_Ratio = Column(REAL)
    NonFDLT_Ratio = Column(REAL)
    NSED = Column(REAL)
    CB_Dcc = Column(REAL)
    SettVolP_mYr = Column(REAL)
    SettVolN_mYr = Column(REAL)
    K_mmHr = Column(REAL)
    CHLAR = Column(REAL)
    SECCIR = Column(REAL)
    InitialVolume = Column(REAL)
    InitialSediment_mgL = Column(REAL)
    InitialNO3_mgL = Column(REAL)
    InitialSolP_mgL = Column(REAL)
    InitialOrgN_mgL = Column(REAL)
    InitialOrgP_mgL = Column(REAL)

    def __init__(self, id, subbasin, receive_reach):
        
        self.ID = id
        self.Subbasin = subbasin
        self.receive_reach_id = receive_reach
        
        self.OperationYear = 0
        self.MaxL_m = 100
        self.MaxW_m = 50
        self.MaxD_m = 10
        self.SideSlope = 1.5
        self.EndSlope = 4
        self.MaxVolume_104M3 = 2.3
        self.FDLT_Ratio = 1
        self.NonFDLT_Ratio = 0
        self.NSED = 5
        self.CB_Dcc = 0.184
        self.SettVolP_mYr = 10
        self.SettVolN_mYr = 10
        self.K_mmHr = 0.05
        self.CHLAR = 1
        self.SECCIR = 1
        self.InitialVolume = 100
        self.InitialSediment_mgL = 5
        self.InitialNO3_mgL = 0.5
        self.InitialSolP_mgL = 0.05
        self.InitialOrgN_mgL = 0.5
        self.InitialOrgP_mgL = 0.05



from whitebox_workflows import WbEnvironment, Raster, RasterDataType, Vector
import numpy as np
import math
from datetime import datetime

class Delineation:   
    def __init__(self):
        pass

    # def getRasterCellAreaM2(raster:Raster):
    #     """Calculate the cell area in km2 assuming the resolution is in m"""
    #     #return raster.configs.resolution_x * raster.configs.resolution_y
    #     pass

    # def getRasterCellAreaKM2(raster:Raster):
    #     """Calculate the cell area in km2 assuming the resolution is in m"""
    #     #return getRasterCellAreaM2(raster) / 1e6
        

    # def calculateStreamDimension(flowAcc:Raster, mask:Raster, parameterA: float, parameterB: float) -> Raster:
    #     """
        
    #     Calculate Stream Width/Depth as (FlowAcc * Cell Area * A)^B. Width and depth will be calculated using different parameters. 
        
    #     Parameters:
    #     flowAcc         : Flow Accumualation Raster
    #     mask            : Stream Network Raster
    #     parameterA   : A
    #     parameterB   : B
        
    #     """

    #     return (flowAcc * getRasterCellArea(mask) * parameterA) ** parameterB

    # def lookupValues(inputRaster:Raster, mask: Raster, lookup:dict) -> Raster:
    #     """
    #     Replace the raster values based on lookup table. The application could be

    #     1. generate field capacity raster based on soil lookup table
    #     2. generate manning raster based on landuse lookup table
    #     3. 
        
    #     Parameters:
    #     inputRaster         : the input raster
    #     mask                : mask raster
    #     lookup              : lookup table assuming the input value is the key and output value is the value   
    #     """

    #     pass

    # def getSlopeRadius(slopeDeg:Raster)->Raster:
    #     return (slopeDeg * math.pi / 180).tan()

    # def calculateReachVelocity(manning:Raster, slopeRadius:Raster, streamDepth:Raster, mask:Raster, min:float = 0.005, max: float = 3) -> Raster:
    #     """
    #     Calculate full reach velocity using Manning equation

    #     Parameters:
    #     manning             : Manning raster
    #     slopeDeg            : Slope raster in degree. Could use radius instead
    #     streamDepth         : Stream depth raster
    #     mask                : mask raster
    #     min                 : min velocity
    #     max                 : max velocity

    #     """
    #     velocity = slopeRadius.sqrt() * streamDepth**0.6667 / manning
    #     return velocity.min(max).max(min)

    # def calculateWetnessIndex(flowAcc:Raster, slopeRadius:Raster, mask:Raster)->Raster:
    #     """
    #     Calculate wetness index using flow accumulation and slope
    #     """

    #     return (flowAcc * getRasterCellAreaM2(mask)).log() / slopeRadius.tan()

    # def calculateIntialSoilMoisture(wetnessIndex:Raster, fieldCapacity:Raster, mask:Raster, minSaturation:float=0.05, maxSaturation:float=1)->Raster:
    #     """
    #     Calculate the initial soil moisture using linear interpolation
    #     """

    #     minWetnessIndex = wetnessIndex.configs.minimum
    #     maxWetnessIndex = wetnessIndex.configs.maximum * 0.8

    #     wti = wetnessIndex.max(maxWetnessIndex)
    #     ratio = (wti - minWetnessIndex) * (maxSaturation - minSaturation) / (maxWetnessIndex - minWetnessIndex) + minSaturation
    #     return ratio * fieldCapacity
    
    def Raise_watershed(watershed_raster:Raster):
        """
        Raise the height of the watershed's extent
        """
        return watershed_raster.con("value > 0", watershed_raster + 200, watershed_raster.configs.nodata)

    def ValueAccumD8(self, d8_ras_file, value_ras_file,output_ras_file, is_average):
        row, col, x, y = 0, 0, 0, 0
        z, z2 = 0.0, 0.0
        v, v2 = 0.0, 0.0
        i = 0
        dX = [1, 1, 1, 0, -1, -1, -1, 0]
        dY = [-1, 0, 1, 1, 1, 0, -1, -1]
        inflowingVals = [16, 32, 64, 128, 1, 2, 4, 8]
        numInNeighbours = 0.0
        isAvg = False
        flag = False
        flowDir = 0.0

        isAvg = is_average

        try:
            wbe = WbEnvironment()

            pntr = wbe.read_raster(d8_ras_file)
            valueRas = wbe.read_raster(value_ras_file)

            noData = pntr.configs.nodata
            rows = pntr.configs.rows
            cols = pntr.configs.columns
            rowsLessOne = rows - 1
    
            output = wbe.new_raster(pntr.configs)
            temp = np.full((rows, cols), 1.0)

            tmpGrid1 = wbe.new_raster(pntr.configs)

            for row in range(rows):
                for col in range(cols):
                    if pntr[row, col] != noData and valueRas[row, col] != noData:
                        z = 0
                        for i in range(8):
                            if pntr[row + dY[i], col + dX[i]] == inflowingVals[i]:
                                z += 1
                        tmpGrid1[row, col] = z
                        output[row, col] = valueRas[row, col]
                    else:
                        temp[row, col] = noData

            for row in range(rows):
                for col in range(cols):
                    if tmpGrid1[row, col] == 0:
                        tmpGrid1[row, col] = -1
                        x, y = col, row
                        while True:
                            z = temp[y, x]
                            v = output[y, x]
                            flowDir = pntr[y, x]
                            if flowDir > 0:
                                i = int(math.log(flowDir) / math.log(2))
                                x += dX[i]
                                y += dY[i]
                                z2 = temp[y, x]
                                temp[y, x] = z2 + z
                                v2 = output[y, x]
                                output[y, x] = v2 + v
                                numInNeighbours = tmpGrid1[y, x] - 1
                                tmpGrid1[y, x] = numInNeighbours
                                if numInNeighbours == 0:
                                    tmpGrid1[y, x] = -1
                                    flag = True
                                else:
                                    flag = False
                            else:
                                flag = False
                            if not flag:
                                break

            if isAvg:
                for row in range(rows):
                    for col in range(cols):
                        flowDir = output[row, col]
                        if flowDir != noData and valueRas[row, col] != noData:
                            output[row, col] = output[row, col] / temp[row, col]
                        else:
                            output[row, col] = noData

            #save here
            wbe.write_raster(output, output_ras_file)

        except MemoryError:
            print("memory error")
        except Exception as e:
            print(e)
        finally:
            pass

    @staticmethod
    def get_flow_path_length(flow_dir_raster:Raster, dem_raster:Raster)->Raster:       
        """
        Replace plugin MaxUpslopeFlowpathLength
        """

        row, col, x, y = 0, 0, 0, 0
        z = 0.0
        currentVal = 0.0
        i = 0
        dX = [1, 1, 1, 0, -1, -1, -1, 0]
        dY = [-1, 0, 1, 1, 1, 0, -1, -1]
        inflowingVals = [16, 32, 64, 128, 1, 2, 4, 8]
        LnOf2 = 0.693147180559945
        numInNeighbours = 0.0
        flag = False
        flowDir = 0.0
        flowLength = 0.0
        
        
        rows = flow_dir_raster.configs.rows
        cols = flow_dir_raster.configs.columns
        noData = flow_dir_raster.configs.nodata
        gridResX = flow_dir_raster.configs.resolution_x
        gridResY = flow_dir_raster.configs.resolution_y

        diagGridRes = np.sqrt(gridResX * gridResX + gridResY * gridResY)
        gridLengths = [diagGridRes, gridResX, diagGridRes, gridResY, diagGridRes, gridResX, diagGridRes, gridResY]

        wbe = WbEnvironment()  
        flow_length_raster = wbe.new_raster(dem_raster.configs)
        for row in range(rows):
            for col in range(cols):
                flow_length_raster[row, col] = 0
        temp_raster = wbe.new_raster(dem_raster.configs)
        
        for row in range(rows):
            for col in range(cols):
                if flow_dir_raster[row, col] != noData:
                    z = 0
                    for i in range(8):
                        if flow_dir_raster[row + dY[i], col + dX[i]] == inflowingVals[i]:
                            z += 1
                    temp_raster[row, col] = z
                else:
                    flow_length_raster[row, col] = noData

        for row in range(rows):
            for col in range(cols):
                if temp_raster[row, col] == 0:
                    temp_raster[row, col] = -1
                    flag = False
                    x = col
                    y = row
                    while True:
                        flowLength = flow_length_raster[y, x]
                        flowDir = flow_dir_raster[y, x]
                        if flowDir > 0:
                            i = int(np.log(flowDir) / LnOf2)
                            flowLength += gridLengths[i]
                            x += dX[i]
                            y += dY[i]
                            
                            currentVal = flow_length_raster[y, x]
                            if flowLength > currentVal:
                                flow_length_raster[y, x] = flowLength
                            
                            numInNeighbours = temp_raster[y, x] - 1
                            temp_raster[y, x] = numInNeighbours
                            
                            if numInNeighbours == 0:
                                temp_raster[y, x] = -1
                                flag = True
                            else:
                                flag = False
                        else:
                            flag = False
                        if not flag:
                            break
        
        return flow_length_raster

    @staticmethod
    def generate_subbasin(flow_dir_raster:Raster, outlet_raster:Raster)->Raster:
        """
        The WhiteBox watershed plugin. 
        Just for testing purpose as the subbasin generated by WhiteBox version is slightly different than the one geneted with the worflow version.
        """
        row, col, x, y = 0, 0, 0, 0
        z = 0
        i, c = 0, 0
        dX = [1, 1, 1, 0, -1, -1, -1, 0]
        dY = [-1, 0, 1, 1, 1, 0, -1, -1]
        flag = False
        flowDir = 0
        outletID = 0

        wbe = WbEnvironment()
        LnOf2 = 0.693147180559945
        rows = flow_dir_raster.configs.rows
        cols = flow_dir_raster.configs.columns
        flow_dir_nodata = flow_dir_raster.configs.nodata
        outlet_nodata = outlet_raster.configs.nodata
        subbasin_raster = wbe.new_raster(outlet_raster.configs)
        subbasin_nodata = subbasin_raster.configs.nodata

        #output = np.full_like(pntr, -999)

        for row in range(rows):
            for col in range(cols):
                z = outlet_raster[row, col]
                if z != 0 and z != outlet_nodata:
                    subbasin_raster[row, col] = z

        for row in range(rows):
            for col in range(cols):
                if subbasin_raster[row, col] == subbasin_nodata and outlet_raster[row, col] != subbasin_nodata:
                    flag = False
                    x = col
                    y = row
                    while not flag:
                        flowDir = flow_dir_raster[y, x]
                        if flowDir > 0:
                            c = int(np.log(flowDir) / LnOf2)
                            x += dX[c]
                            y += dY[c]
                            z = subbasin_raster[y, x]
                            if z != subbasin_nodata:
                                outletID = z
                                flag = True
                        else:
                            outletID = subbasin_nodata
                            flag = True

                    flag = False
                    x = col
                    y = row
                    subbasin_raster[y, x] = outletID
                    while not flag:
                        flowDir = flow_dir_nodata[y, x]
                        if flowDir > 0:
                            c = int(np.log(flowDir) / LnOf2)
                            x += dX[c]
                            y += dY[c]
                            z = subbasin_raster[y, x]
                            if z != subbasin_nodata:
                                flag = True
                        else:
                            flag = True
                        subbasin_raster[y, x] = outletID
                elif flow_dir_raster[row, col] != flow_dir_nodata:
                    subbasin_raster[row, col] = subbasin_nodata
 
        return subbasin_raster


    @staticmethod
    def get_pour_points(stream_network_raster:Raster, flow_dir_raster:Raster, structure_rasters:list)->Vector:
        """
        * This method generate the pour points of the stream with the option of
        * setting exception areas, where pour points will be removed.
        *
        * @param streamHeader Input stream raster file. All cells pouring to the
        * same stream cell will be classify as a pour point
        * @param flowDirHeader Flow direction raster file path. This file will be
        * used to tracking the flow direction of the stream and find the pour
        * points
        * @param outputHeader Pour point raster file path (Output file).
        * @param exceptionAreas Array of exception areas (Optional). All pour
        * points within these areas will be removed

        This was used when uer click the watershed outlet button to provide user a list of outlets for selection. 
        This will not be used assuming provides the outlets in a shapefile. 

        Replace buildPourPoints
        """

        dX = [1, 1, 1, 0, -1, -1, -1, 0]
        dY = [-1, 0, 1, 1, 1, 0, -1, -1]
        inflowing_vals = [16, 32, 64, 128, 1, 2, 4, 8]

        rows = flow_dir_raster.configs.rows
        cols = flow_dir_raster.configs.columns
        no_data = flow_dir_raster.configs.nodata
        has_exception = len(structure_rasters) > 0

        wbe = WbEnvironment()
        configs = flow_dir_raster.configs
        configs.data_type = RasterDataType.F32
        stream_outlets_raster = wbe.new_raster(configs)

        count = 0
        for row in range(rows):
            for col in range(cols):
                if stream_network_raster[row, col] > 0:
                    flow_dir = flow_dir_raster[row, col]
                    if flow_dir == 0:  # End of stream network
                        count += 1
                        stream_outlets_raster[row, col] = count
                        break

                    # Check if it is a headwater location
                    num_neighbouring_stream_cells = 0
                    for c in range(8):
                        x = col + dX[c]
                        y = row + dY[c]
                        if stream_network_raster[y, x] > 0 and flow_dir_raster[y, x] == inflowing_vals[c]:
                            num_neighbouring_stream_cells += 1

                    if num_neighbouring_stream_cells > 1:
                        if has_exception:
                            flag = True
                            for exception in structure_rasters:
                                if exception[row, col] > 0:
                                    flag = False
                                    break
                            if not flag:
                                break

                        for c in range(8):
                            x = col + dX[c]
                            y = row + dY[c]
                            if stream_network_raster[y, x] > 0 and flow_dir_raster[y, x] == inflowing_vals[c]:
                                if has_exception:
                                    flag = True
                                    for exception in structure_rasters:
                                        if exception[y, x] > 0:
                                            flag = False
                                            break
                                    if not flag:
                                        break

                                count += 1
                                stream_outlets_raster[y, x] = count

        return wbe.raster_to_vector_points(stream_outlets_raster)

    @staticmethod
    def build_stream_network_link_to_outlets(stream_network_raster:Raster, flow_dir_raster:Raster, outlets_raster:Raster)->Raster:
        """
        it basically add small stream that connects to outlets

        replace buildStreamNetworkLink2Outlets
        """
        #dX = [1, 1, 1, 0, -1, -1, -1, 0]
        #dY = [-1, 0, 1, 1, 1, 0, -1, -1]
        #dmove = [(-1, 1), (0, 1), (1, 1), (1, 0), (1, -1), (0, -1), (-1, -1), (-1, 0)]
        dmove_dic = {1: (-1, 1), 2: (0, 1), 4: (1, 1), 8: (1, 0), 16: (1, -1), 32: (0, -1), 64: (-1, -1), 128: (-1, 0)}  

        wbe = WbEnvironment()
        rows = stream_network_raster.configs.rows
        cols = stream_network_raster.configs.columns
        no_data = stream_network_raster.configs.nodata

        stream_network_modified_raster = wbe.new_raster(stream_network_raster.configs)
        flag_raster = wbe.new_raster(stream_network_raster.configs)

        for row in range(rows):
            for col in range(cols):
                #skip if the cell is already handled
                if flag_raster[row, col] > 0:
                    continue

                if stream_network_raster[row, col] > 0:
                    stream_network_modified_raster[row, col] = stream_network_raster[row, col]
                    flag_raster[row, col] = 1
                elif outlets_raster[row, col] > 0:
                    if flow_dir_raster[row, col] < 0:
                        break

                    x = col
                    y = row
                    value = no_data

                    path = []
                    path.append((x,y))
                    while value < 0:
                        flow_dir = flow_dir_raster[y, x]
                        if flow_dir > 0:
                            mv = dmove_dic[flow_dir]

                            x += mv[1]
                            y += mv[0]
                            value = stream_network_raster[y, x]
                            path.append((x,y))
                        else:
                            break

                    if value < 0:
                        value = 1

                    for point in path:
                        stream_network_modified_raster[point[1],point[0]] = value
                        flag_raster[point[1],point[0]] = 1

        return stream_network_modified_raster

    @staticmethod
    def reorder_raster_id(raster:Raster):
        """
        it will modify the raster

        replace reorderRasterID
        """

        rows = raster.configs.rows
        cols = raster.configs.columns

        count = 0
        id_map = {}
        for row in range(rows):
            for col in range(cols):
                id = int(raster[row, col])
                if id > 0:
                    if id not in id_map:
                        count += 1
                        id_map[id] = count
                    raster[row, col] = id_map[id]






from ...delineation.structure_attribute import StructureAttribute

class Dugout:
    """Parameter Table for BMP: Dugout (38)"""

    def __init__(self, attribute:StructureAttribute = None):
        self.Scenario = -1
        """Dugout ID, obtained from dugout input data"""
        self.ID = attribute.id
        """Dugout type, 0 – filled with runoff, 1 – connected to a water course"""
        self.Type = 0
        """Operation starting year (designed for setting up dugout scenarios)"""
        self.Year = 0
        """Subbasin number, obtained from the watershed delineation"""
        self.Subbasin = attribute.subbasin 
        """Length of dugout when filled (m)"""
        self.MaxL_m = 100
        """Width of dugout when filled (m)"""
        self.MaxW_m = 50
        """Dugout depth when filled (m)"""
        self.MaxD_m = 10
        """Dugout side slope (-)"""
        self.SideSlope = 1.5
        """Dugout end slope (-)"""
        self.EndSlope = 4
        """Dugout volume capacity (104m3), calculated from above 4 parameters"""
        self.MaxVolume_104M3 = 0
        """Dugout drainage area (ha), obtained from dugout drainage area distribution"""
        self.Drainage_Area = attribute.contribution_area
        """Animal ID"""
        self.AniID = 2
        """Adult animal number"""
        self.AniAdult = 20
        """Non-adult animal number"""
        self.AniNonAdult = 20
        """Normal (equilibrium) sediment concentration in wetland (mg/l)"""
        self.NSED = 5
        """Sediment settling decay constant (1/day)"""
        self.Dcc = 0.184
        """Median partical size"""
        self.D50 = 10
        """Phosphorous settling rate in catch basin (m/year)"""
        self.SettVolP_mYr = 10
        """Nitrogen settling rate in catch basin (m/year)"""
        self.SettVolN_mYr = 10
        """Hydraulic conductivity of catch basin bottom (mm/hr)"""
        self.K_mmHr = 0.05
        """Chlorophyll production coefficient of catch basin"""
        self.CHLAR = 1
        """Water clarity coefficient of catch basin (m)"""
        self.SECCIR = 1
        """Initial water volume in catch basin (% of capacity),"""
        self.InitialVolume = 100
        """Initial sediment concentration in catch basin (mg/l)"""
        self.InitialSediment_mgL = 5
        """Initial concentration of NO3-N in catch basin (mg/l)"""
        self.InitialNO3_mgL = 0.5
        """Initial concentration of soluble P in catch basin (mg/l)"""
        self.InitialSolP_mgL = 0.05
        """Initial concentration of organic N in catch basin (mg/l)"""
        self.InitialOrgN_mgL = 0.5
        """Initial concentration of organic P in catch basin (mg/l)"""
        self.InitialOrgP_mgL = 0.05



from typing import Any
from sqlalchemy import Column, Integer, TEXT, REAL
from .bmp_table import BMPTable
from .bmp_management_base import BMPManagementBaseWithYear
from ...names import Names

class ManureFeedlot(BMPTable):
    """Parameter Table for BMP: Manure feedlot (29)"""
    __tablename__ = Names.bmp_table_name_manure_feed_lot_parameter
    """ID"""
    ID = Column(Integer, primary_key = True)
    """Name"""
    Name = Column(TEXT)
    """Description"""
    Description = Column(TEXT)
    """Producer ID. Not used in the module"""
    ProducerID = Column(Integer)
    """Subbasin. Not used in the module"""
    Subbasin = Column(Integer)
    """The id of animal whose parameter could be found in livestok_parameter table"""
    AniID = Column(Integer)
    """The id of catch basin which the feedlot would drains to."""
    CatBasID = Column(Integer)
    """Initial manure storage of the feedlot"""
    ManInitial = Column(REAL)
    """Average CN increase fraction compare to non-feedlot area"""
    CN_change = Column(REAL)
    """Average PRC increase fraction compare to non-feedlot area"""
    PRC_change = Column(REAL)
    """Manure event mean concentration"""
    Manure_EMC = Column(REAL)

    def __init__(self, id:int, animal_id:int, subbasin:int, catch_basin_id:int):
        self.ID = id
        self.ProducerID = self.ID
        self.Subbasin = subbasin
        self.CatBasID = catch_basin_id
        self.AniID = animal_id
        self.ManInitial = 0
        self.CN_change = 0.2
        self.PRC_change = 0.2
        self.Manure_EMC = 100000

class ManureFeedlotManagement(BMPManagementBaseWithYear):
    """
    Distribution Table for BMP: Manure feedlot (29)
    It has two operations:
    1. Accumulate: FDLMon,FDLDay,Days,AniAdult,AniNonAdult
    2. Remove: ManStoID,ManStoDis,ManRemMon,ManRemDay,ManRemFra
    """
    
    def __init__(self, feedlot_id:int, adult:int, non_adult:int, manure_stroage_id:str, manure_storage_distribution:str):
        """All data will be read from feedlot boundary shapefile"""
        super().__init__()

        self.Location = feedlot_id

        self.FDLMon = 1
        self.FDLDay = 1
        self.Days = 120

        """Number of adult animals"""
        self.AniAdult = adult
        """Number of non-adult animals"""
        self.AniNonAdult = non_adult

        """Manure storage id spearate by /"""
        self.ManStoID = manure_stroage_id

        """
        Manure storage distribution ratio separated by /
        The total should be <= 1
        The number of ratio is same as the number of manure storage        
        """
        self.ManStoDis = manure_storage_distribution
        self.ManRemMon = 10
        self.ManRemDay = 31
        self.ManRemFra = 1

    @staticmethod
    def column_types()->dict:
        feedlot = ManureFeedlotManagement(0,0,0,"","")
        return {col:(REAL if col == "ManRemFra" else (TEXT if col in ["ManStoID","ManStoDis"] else Integer)) for col in dir(feedlot) if "__" not in col}



from whitebox_workflows import Raster, Vector
from ..vector_extension import VectorExtension
from ..raster_extension import RasterExtension
from ..database.bmp.bmp_scenarios import BMP_scenarios
from ..names import Names
from .bmp_type import BMPParameters, BMPDistributions
import pandas as pd

class BMP:
    def __init__(self,bmp_vector:Vector, subbasin_raster:Raster,):
        self.bmp_vector = bmp_vector
        self.subbasin_raster = subbasin_raster

        self.__ids = None
        self.__subbasins = None
        self.__bmp_raster = None
    
    @property
    def ids(self)->list:
        """unique ids"""
        if self.__ids is None:
            self.__ids = VectorExtension.get_unique_ids(self.bmp_vector)

        return self.__ids
    
    @property
    def bmp_raster(self)->Raster:
        if self.__bmp_raster is None:
            self.__bmp_raster = VectorExtension.vector_to_raster(self.bmp_vector, self.subbasin_raster)

        return self.__bmp_raster

    @property
    def subbasins(self)->dict:
        """Get id and subbasin dictionary"""
        if self.__subbasins is None:
            self.__subbasins = RasterExtension.get_zonal_statistics(self.subbasin_raster, self.bmp_raster,"max","subbasin")["subbasin"].to_dict()

        return self.__subbasins
    
    @staticmethod
    def generate_bmp_scenarios_df(bmp_types:list)->pd.DataFrame:
        bmps = [BMP_scenarios(type, BMPDistributions[type], BMPParameters[type]) for type in bmp_types]
        return pd.DataFrame([vars(rb) for rb in bmps])
        













    

    